#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
__all__ = (
    "EscalatorsRangeV1",
    "ElevatorsRangeV1",
    "ConveyingEquipmentRangeV1",
)

# NB! This is a generated code. Do not edit it manually. Please see src/openepd/model/specs/README.md


from openepd.compat.pydantic import pyd
from openepd.model.specs.base import BaseOpenEpdHierarchicalSpec
from openepd.model.specs.enums import ElevatorsBuildingRise, ElevatorsUsageIntensity
from openepd.model.validation.quantity import (
    AmountRangeCapacityPerHour,
    AmountRangeLengthMm,
    AmountRangeMass,
    AmountRangeSpeed,
)


class EscalatorsRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Escalators category performance specification.

    A moving staircase consisting of a circulating belt of steps driven by a motor,
    which conveys people between the floors of a building.

    Range version.
    """

    _EXT_VERSION = "1.1"

    vertical_rise: AmountRangeLengthMm | None = pyd.Field(
        default=None, description="The vertical distance between the top and bottom landings of an escalator"
    )
    speed: AmountRangeSpeed | None = pyd.Field(default=None, description="Reference speed of the escalator")
    step_width: AmountRangeLengthMm | None = pyd.Field(default=None, description="Width of the escalator steps")
    max_capacity: AmountRangeCapacityPerHour | None = pyd.Field(
        default=None, description="Max capacity of escalator in persons per hour"
    )
    indoor: bool | None = pyd.Field(default=None, description="Escalator can be used for indoor applications")
    outdoor: bool | None = pyd.Field(default=None, description="Escalator can be used for outdoor applications")


class ElevatorsRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Car that moves in a vertical shaft to carry passengers or freight between the levels of a multistory building.

    Range version.
    """

    _EXT_VERSION = "1.0"

    usage_intensity: list[ElevatorsUsageIntensity] | None = pyd.Field(default=None, description="")
    travel_length: AmountRangeLengthMm | None = pyd.Field(default=None, description="")
    rated_load: AmountRangeMass | None = pyd.Field(default=None, description="")
    rated_speed: AmountRangeSpeed | None = pyd.Field(default=None, description="")
    building_rise: list[ElevatorsBuildingRise] | None = pyd.Field(default=None, description="")


class ConveyingEquipmentRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Conveying Equipment.

    Range version.
    """

    _EXT_VERSION = "1.1"

    Elevators: ElevatorsRangeV1 | None = None
    Escalators: EscalatorsRangeV1 | None = None



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
__all__ = (
    "AluminiumSidingRangeV1",
    "SteelSidingRangeV1",
    "ZincSidingRangeV1",
    "ShingleAndShakeSidingRangeV1",
    "MetalSidingRangeV1",
    "CompositionSidingRangeV1",
    "FiberCementSidingRangeV1",
    "InsulatedVinylSidingRangeV1",
    "PlywoodSidingRangeV1",
    "PolypropyleneSidingRangeV1",
    "SolidWoodSidingRangeV1",
    "VinylSidingRangeV1",
    "SidingRangeV1",
    "InsulatedRoofPanelsRangeV1",
    "InsulatedWallPanelsRangeV1",
    "RoofPanelsRangeV1",
    "StoneCladdingRangeV1",
    "WallPanelsRangeV1",
    "CladdingRangeV1",
)

# NB! This is a generated code. Do not edit it manually. Please see src/openepd/model/specs/README.md


from openepd.compat.pydantic import pyd
from openepd.model.specs.base import BaseOpenEpdHierarchicalSpec
from openepd.model.specs.enums import CladdingFacingMaterial, CladdingInsulatingMaterial, SidingFormFactor
from openepd.model.validation.quantity import AmountRangeLengthMm, AmountRangeRValue


class AluminiumSidingRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Exterior siding product made primarily from aluminium.

    Range version.
    """

    _EXT_VERSION = "1.0"


class SteelSidingRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Exterior siding product made primarily from steel.

    Range version.
    """

    _EXT_VERSION = "1.0"


class ZincSidingRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Exterior siding product made primarily from zinc.

    Range version.
    """

    _EXT_VERSION = "1.0"


class ShingleAndShakeSidingRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Shingle & shake siding.

    Range version.
    """

    _EXT_VERSION = "1.0"


class MetalSidingRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Exterior siding product made of metal such as steel, aluminum, etc.

    Range version.
    """

    _EXT_VERSION = "1.0"

    AluminiumSiding: AluminiumSidingRangeV1 | None = None
    SteelSiding: SteelSidingRangeV1 | None = None
    ZincSiding: ZincSidingRangeV1 | None = None


class CompositionSidingRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Composite wood siding composed of wood wafers and resin.

    Range version.
    """

    _EXT_VERSION = "1.0"


class FiberCementSidingRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Composite siding product made of cement and cellulose fibers.

    Range version.
    """

    _EXT_VERSION = "1.0"


class InsulatedVinylSidingRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Vinyl cladding product integrated with manufacturer-installed insulation.

    Range version.
    """

    _EXT_VERSION = "1.0"

    thickness: AmountRangeLengthMm | None = pyd.Field(default=None, description="")


class PlywoodSidingRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Siding made of plywood boards.

    Range version.
    """

    _EXT_VERSION = "1.0"


class PolypropyleneSidingRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Exterior wall cladding made from polypropylene, which may contain fillers or reinforcements.

    Range version.
    """

    _EXT_VERSION = "1.0"


class SolidWoodSidingRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Siding products made of wood including shingle & shake siding.

    Range version.
    """

    _EXT_VERSION = "1.0"

    ShingleAndShakeSiding: ShingleAndShakeSidingRangeV1 | None = None


class VinylSidingRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Exterior wall cladding made principally from rigid polyvinyl chloride (PVC).

    Range version.
    """

    _EXT_VERSION = "1.0"

    thickness: AmountRangeLengthMm | None = pyd.Field(default=None, description="")


class SidingRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Long narrow products for exterior wall face of building.

    Typically made of, e.g., metal, solid wood, plywood, plastic, composition, fiber cement, etc.

    Range version.
    """

    _EXT_VERSION = "1.0"

    insulated: bool | None = pyd.Field(default=None, description="")
    ventilated: bool | None = pyd.Field(default=None, description="")
    paint_or_stain_required: bool | None = pyd.Field(default=None, description="")
    r_value: AmountRangeRValue | None = pyd.Field(default=None, description="")
    form_factor: list[SidingFormFactor] | None = pyd.Field(default=None, description="")
    MetalSiding: MetalSidingRangeV1 | None = None
    CompositionSiding: CompositionSidingRangeV1 | None = None
    FiberCementSiding: FiberCementSidingRangeV1 | None = None
    InsulatedVinylSiding: InsulatedVinylSidingRangeV1 | None = None
    PlywoodSiding: PlywoodSidingRangeV1 | None = None
    PolypropyleneSiding: PolypropyleneSidingRangeV1 | None = None
    SolidWoodSiding: SolidWoodSidingRangeV1 | None = None
    VinylSiding: VinylSidingRangeV1 | None = None


class InsulatedRoofPanelsRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Insulated roof panels performance specification.

    Range version.
    """

    _EXT_VERSION = "1.0"

    r_value: AmountRangeRValue | None = pyd.Field(default=None, description="")
    insulating_material: list[CladdingInsulatingMaterial] | None = pyd.Field(default=None, description="")


class InsulatedWallPanelsRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Insulated wall panels performance specification.

    Range version.
    """

    _EXT_VERSION = "1.0"

    r_value: AmountRangeRValue | None = pyd.Field(default=None, description="")
    insulating_material: list[CladdingInsulatingMaterial] | None = pyd.Field(default=None, description="")


class RoofPanelsRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Roof panels performance specification.

    Range version.
    """

    _EXT_VERSION = "1.0"


class StoneCladdingRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Stone cladding performance specification.

    Range version.
    """

    _EXT_VERSION = "1.0"


class WallPanelsRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Wall panels performance specification.

    Range version.
    """

    _EXT_VERSION = "1.0"


class CladdingRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Cladding performance specification.

    Range version.
    """

    _EXT_VERSION = "1.0"

    thickness: AmountRangeLengthMm | None = pyd.Field(default=None, description="")
    facing_material: list[CladdingFacingMaterial] | None = pyd.Field(default=None, description="")
    Siding: SidingRangeV1 | None = None
    InsulatedRoofPanels: InsulatedRoofPanelsRangeV1 | None = None
    InsulatedWallPanels: InsulatedWallPanelsRangeV1 | None = None
    RoofPanels: RoofPanelsRangeV1 | None = None
    StoneCladding: StoneCladdingRangeV1 | None = None
    WallPanels: WallPanelsRangeV1 | None = None



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from typing import Annotated, Optional

from openlocationcode import openlocationcode

from openepd.compat.pydantic import pyd
from openepd.model.base import BaseOpenEpdSchema
from openepd.model.common import Location, WithAltIdsMixin, WithAttachmentsMixin
from openepd.model.validation.common import ReferenceStr


class OrgRef(BaseOpenEpdSchema):
    """Represents Organisation with minimal data."""

    web_domain: str | None = pyd.Field(
        description="A web domain owned by organization. Typically is the org's home website address", default=None
    )
    name: str | None = pyd.Field(
        max_length=200,
        description="Common name for organization",
        example="C Change Labs",
        default=None,
    )
    ref: ReferenceStr | None = pyd.Field(
        default=None,
        example="https://openepd.buildingtransparency.org/api/orgs/c-change-labs.com",
        description="Reference to this Org's JSON object",
    )


class Org(WithAttachmentsMixin, WithAltIdsMixin, OrgRef):
    """Represent an organization."""

    alt_names: Annotated[list[str], pyd.conlist(pyd.constr(max_length=200), max_items=255)] | None = pyd.Field(
        description="List of other names for organization",
        example=["C-Change Labs", "C-Change Labs inc."],
        default=None,
    )
    # TODO: NEW field, not in the spec

    owner: Optional["OrgRef"] = pyd.Field(description="Organization that controls this organization", default=None)
    subsidiaries: Annotated[list["OrgRef"], pyd.conlist(pyd.constr(max_length=200), max_items=255)] | None = pyd.Field(
        description="Organizations controlled by this organization",
        default=None,
    )
    hq_location: Location | None = pyd.Field(
        default=None,
        description="Location of a place of business, preferably the corporate headquarters.",
    )


class PlantRef(BaseOpenEpdSchema):
    """Represents Plant with minimal data."""

    id: str | None = pyd.Field(
        description="Plus code (aka Open Location Code) of plant's location and "
        "owner's web domain joined with `.`(dot).",
        example="865P2W3V+3W.interface.com",
        default=None,
    )
    name: str | None = pyd.Field(
        max_length=200,
        description="Manufacturer's name for plant. Recommended < 40 chars",
        example="Dalton, GA",
        default=None,
    )
    ref: ReferenceStr | None = pyd.Field(
        default=None,
        example="https://openepd.buildingtransparency.org/api/orgs/c-change-labs.com",
        description="Reference to this Plant's JSON object",
    )


class Plant(PlantRef, WithAttachmentsMixin, WithAltIdsMixin):
    """Represent a manufacturing plant."""

    pluscode: str | None = pyd.Field(
        default=None,
        description="(deprecated) Plus code (aka Open Location Code) of plant's location",
        deprecated="Pluscode field is deprecated. If users need a pluscode they can obtain it from "
        "`id` like this: `id.spit('.', maxsplit=1)[0]`",
    )
    latitude: float | None = pyd.Field(
        default=None, description="(deprecated) Latitude of the plant location. Use 'location' fields instead."
    )
    longitude: float | None = pyd.Field(
        default=None, description="(deprecated) Longitude of the plant location. Use 'location' fields instead."
    )
    owner: Org | None = pyd.Field(description="Organization that owns the plant", default=None)
    address: str | None = pyd.Field(
        max_length=200,
        default=None,
        description="(deprecated) Text address, preferably geocoded. Use 'location' fields instead",
        example="1503 Orchard Hill Rd, LaGrange, GA 30240, United States",
    )
    contact_email: pyd.EmailStr | None = pyd.Field(
        description="Email contact", example="info@interface.com", default=None
    )
    location: Location | None = pyd.Field(description="Location of the plant", default=None)

    @classmethod
    def get_asset_type(cls) -> str | None:
        """Return the asset type of this class (see BaseOpenEpdSchema.get_asset_type for details)."""
        return "org"

    @pyd.validator("id")
    def _validate_id(cls, v: str) -> str:
        if not v:
            return v
        try:
            pluscode, web_domain = v.split(".", maxsplit=1)
        except ValueError as e:
            raise ValueError("Incorrectly formed id: should be pluscode.owner_web_domain") from e

        if not openlocationcode.isValid(pluscode):
            raise ValueError("Incorrect pluscode for plant")

        if not web_domain:
            raise ValueError("Incorrect web_domain for plant")
        return v

    class Config(BaseOpenEpdSchema.Config):
        allow_population_by_field_name = True



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
import abc
import datetime
from typing import Final, Generic, TypeAlias, TypeVar

from openepd.api.dto.base import BaseMetaDto, BaseOpenEpdApiModel, MetaExtensionBase
from openepd.api.dto.meta import PerformanceMetaMixin
from openepd.compat.pydantic import pyd, pyd_generics
from openepd.model.base import AnySerializable, BaseOpenEpdSchema

DEFAULT_PAGE_SIZE: Final[int] = 100
MAX_PAGE_SIZE: Final[int] = 250


class AuditableDto(BaseOpenEpdApiModel, metaclass=abc.ABCMeta):
    """Base class for all DTOs that hold audit information."""

    created_by: str = pyd.Field(
        title="Created By",
        example="johnsmith@cqd.io",
        description="User's email or script name that created this list.",
    )
    updated_by: str = pyd.Field(
        title="Updated By",
        example="bobbuilder@buildingtransparency.org",
        description="User's email or script name that updated this list last time.",
    )
    created_on: datetime.datetime = pyd.Field(
        title="Created On",
        example="2019-06-13T13:17:09+00:00",
        description="A timestamp when this object has been created "
        "in [ISO 8601](https://en.wikipedia.org/wiki/ISO_8601) format.",
    )
    updated_on: datetime.datetime = pyd.Field(
        title="Updated On",
        example="2020-07-13T13:17:09+00:00",
        description="A timestamp when this object has been updated last time "
        "in [ISO 8601](https://en.wikipedia.org/wiki/ISO_8601) format.",
    )


Payload: TypeAlias = AnySerializable
TPayload = TypeVar("TPayload", bound=Payload)
TMetaDto = TypeVar("TMetaDto", bound=BaseMetaDto)
TMetaExtension = TypeVar("TMetaExtension", bound=MetaExtensionBase)


class MetaCollectionDto(BaseOpenEpdApiModel, pyd_generics.GenericModel, Generic[TMetaExtension]):
    """
    This class is intended to be used as a container for different meta objects.

    From a specific controller, you should return a specific subclass of MetaCollectionDto and appropriate mixins. This
    would allow to retain type information to generate schema for meta section of response.

    For example, EPD Search method, which should return some metadata about performance, material filter, and paging,
    would return a concrete subclass of MetaCollectionDto like this one:

    `class EpdSearchViewMeta(MaterialFilterMetaMixin, PagingMetaMixin, WarningMetaMixin, MetaCollectionDto)`

    Mixins are used to define the resulting meta's dictionary-like structure, this would result in a json like:
    ```
    {
    "paging": {"page_size": 100, "total_count": 1, "total_pages": 1},
    "ec3_warnings": null,
    "material_filter": {...}
    }
    ```

    It is generified by the extension type of similar structure (key-dict value), which should not be a part of OpenEPD
    spec. This allows to add custom meta objects to the response, which would be ignored by OpenEPD implementors.
    """

    ext: TMetaExtension | None = None

    class Config(BaseOpenEpdSchema.Config):
        schema_extra = {
            "description": "Base structure of the response meta section",
        }


class BaseMeta(PerformanceMetaMixin, MetaCollectionDto[TMetaExtension], Generic[TMetaExtension], metaclass=abc.ABCMeta):
    """Base class for creating meta objects specific to a controller."""

    pass


TMeta = TypeVar("TMeta", bound=MetaCollectionDto, covariant=True)


class OpenEpdApiResponse(pyd_generics.GenericModel, BaseOpenEpdApiModel, Generic[TPayload, TMeta]):
    """Standard DTO representing response from OpenEPD API server."""

    payload: TPayload
    # there is an issue with using covariant type variables as parameters when used for container mutable types,
    # as described in https://github.com/python/mypy/issues/7049
    # However, in our workflow TMeta is not a container type which is further returned back to caller.
    meta: TMeta  # type: ignore[misc]



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
import abc
from enum import StrEnum
import json
from typing import Any, Callable, ClassVar, Generic, Optional, Type, TypeAlias, TypeVar

from cqd import open_xpd_uuid  # type:ignore[import-untyped,ignore-not-found]

from openepd.compat.pydantic import pyd, pyd_generics
from openepd.model.validation.common import validate_version_compatibility, validate_version_format
from openepd.model.versioning import OpenEpdVersions, Version

AnySerializable: TypeAlias = int | str | bool | float | list | dict | pyd.BaseModel | None
TAnySerializable = TypeVar("TAnySerializable", bound=AnySerializable)

OPENEPD_VERSION_FIELD = "openepd_version"
"""Field name for the openEPD format version."""

OPENAPI_SCHEMA_SERVICE_PROPERTIES = ["ext_version", "ext"]
"""OpenAPI properties which should be moved to the bottom of specification if present. """


class OpenEpdDoctypes(StrEnum):
    """Enum of supported openEPD document types."""

    Epd = "openEPD"
    GenericEstimate = "openGenericEstimate"
    IndustryEpd = "openIndustryEpd"


def modify_pydantic_schema(schema_dict: dict, cls: type) -> dict:
    """
    Modify the schema dictionary to add the required fields.

    :param schema_dict: schema dictionary
    :param cls: class for which the schema was generated
    :return: modified schema dictionary
    """
    for prop_name in OPENAPI_SCHEMA_SERVICE_PROPERTIES:
        prop = schema_dict.get("properties", {}).get(prop_name, None)
        # move to bottom
        if prop is not None:
            del schema_dict["properties"][prop_name]
            schema_dict["properties"][prop_name] = prop

    return schema_dict


class BaseOpenEpdSchema(pyd.BaseModel):
    """Base class for all OpenEPD models."""

    ext: dict[str, AnySerializable] | None = pyd.Field(alias="ext", default=None)

    class Config:
        allow_mutation = True
        validate_assignment = False
        allow_population_by_field_name = True
        use_enum_values = True
        schema_extra: Callable | dict = modify_pydantic_schema

    def to_serializable(self, *args, **kwargs) -> dict[str, Any]:
        """
        Return a serializable dict representation of the DTO.

        It expects the same arguments as the pyd.BaseModel.json() method.
        """
        return json.loads(self.json(*args, **kwargs))

    def has_values(self) -> bool:
        """Return True if the model has any values."""
        return len(self.dict(exclude_unset=True, exclude_none=True)) > 0

    def set_ext(self, ext: "OpenEpdExtension") -> None:
        """Set the extension field."""
        self.set_ext_field(ext.get_extension_name(), ext)

    def set_ext_field(self, key: str, value: AnySerializable) -> None:
        """Add an extension field to the model."""
        if self.ext is None:
            self.ext = {}
        self.ext[key] = value

    def set_ext_field_if_any(self, key: str, value: AnySerializable) -> None:
        """Add an extension field to the model if the value is not None."""
        if value is not None:
            self.set_ext_field(key, value)

    def get_ext_field(self, key: str, default: AnySerializable = None) -> AnySerializable | None:
        """Get an extension field from the model."""
        if self.ext is None:
            return default
        return self.ext.get(key, default)

    def get_typed_ext_field(
        self, key: str, target_type: Type[TAnySerializable], default: Optional[TAnySerializable] = None
    ) -> TAnySerializable:
        """
        Get an extension field from the model and convert it to the target type.

        :raise ValueError: if the value cannot be converted to the target type.
        """
        value = self.get_ext_field(key, default)
        if value is None:
            return None  # type: ignore
        if issubclass(target_type, pyd.BaseModel) and isinstance(value, dict):
            return target_type.parse_obj(value)  # type: ignore[return-value]
        elif isinstance(value, target_type):
            return value
        raise ValueError(f"Cannot convert {value} to {target_type}")

    def get_ext(self, ext_type: Type["TOpenEpdExtension"]) -> Optional["TOpenEpdExtension"]:
        """Get an extension field from the model or None if it doesn't exist."""
        return self.get_typed_ext_field(ext_type.get_extension_name(), ext_type, None)

    def get_ext_or_empty(self, ext_type: Type["TOpenEpdExtension"]) -> "TOpenEpdExtension":
        """Get an extension field from the model or an empty instance if it doesn't exist."""
        return self.get_typed_ext_field(ext_type.get_extension_name(), ext_type, ext_type.construct(**{}))

    @classmethod
    def is_allowed_field_name(cls, field_name: str) -> bool:
        """
        Return True if the field name is defined in the module.

        Both property name and aliases are checked.
        """
        if field_name in cls.__fields__:
            return True
        else:
            for x in cls.__fields__.values():
                if x.alias == field_name:
                    return True
        return False

    @classmethod
    def get_asset_type(cls) -> str | None:
        """
        Return the asset type as it should be written into bundle.

        Only independent (e.g. Pcr, Org, Epd, etc) objects have asset type.
        Supplementary objects (e.g. ResourceUseSet, Location, LatLng) must always return None.
        """
        return None


class BaseOpenEpdGenericSchema(pyd_generics.GenericModel, BaseOpenEpdSchema):
    """Base class for all OpenEPD generic models."""

    pass


class OpenEpdExtension(BaseOpenEpdSchema, metaclass=abc.ABCMeta):
    """Base class for OpenEPD extension models."""

    @classmethod
    @abc.abstractmethod
    def get_extension_name(cls) -> str:
        """Return the name of the extension."""
        pass


TOpenEpdExtension = TypeVar("TOpenEpdExtension", bound=OpenEpdExtension)
TOpenEpdObject = TypeVar("TOpenEpdObject", bound=BaseOpenEpdSchema)
TOpenEpdObjectClass = TypeVar("TOpenEpdObjectClass", bound=Type[BaseOpenEpdSchema])


class RootDocument(abc.ABC, BaseOpenEpdSchema):
    """Base class for all objects representing openEPD root element. E.g. Epd, IndustryEpd, GenericEstimate, etc."""

    _FORMAT_VERSION: ClassVar[str]
    """Version of this document format. Must be defined in the concrete class."""

    doctype: str = pyd.Field(
        description='Describes the type and schema of the document. Must always always read "openEPD".',
        default="openEPD",
    )
    openepd_version: str = pyd.Field(
        description="Version of the document format, related to /doctype",
        default=OpenEpdVersions.get_current().as_str(),
    )

    _version_format_validator = pyd.validator(OPENEPD_VERSION_FIELD, allow_reuse=True, check_fields=False)(
        validate_version_format
    )
    _version_major_match_validator = pyd.validator(OPENEPD_VERSION_FIELD, allow_reuse=True, check_fields=False)(
        validate_version_compatibility("_FORMAT_VERSION")
    )


TRootDocument = TypeVar("TRootDocument", bound=RootDocument)


class BaseDocumentFactory(Generic[TRootDocument]):
    """
    Base class for document factories.

    Extend it to create a factory for a specific document type e.g. for industry epd, epd, etc.
    """

    DOCTYPE_CONSTRAINT: OpenEpdDoctypes
    VERSION_MAP: dict[Version, type[TRootDocument]] = {}

    @classmethod
    def from_dict(cls, data: dict) -> TRootDocument:
        """Create a document from a dictionary."""
        doctype: str | None = data.get("doctype")
        if doctype is None:
            raise ValueError("Doctype not found in the data.")
        if doctype.lower() != cls.DOCTYPE_CONSTRAINT.lower():
            raise ValueError(
                f"Document type {doctype} not supported. This factory supports {cls.DOCTYPE_CONSTRAINT} only."
            )
        version = Version.parse_version(data.get(OPENEPD_VERSION_FIELD, ""))
        for x, doc_cls in cls.VERSION_MAP.items():
            if x.major == version.major:
                if version.minor <= x.minor:
                    return doc_cls(**data)
                else:
                    raise ValueError(
                        f"Unsupported version: {version}. "
                        f"The highest supported version from branch {x.major}.x is {x}"
                    )
        supported_versions = ", ".join(f"{v.major}.x" for v in cls.VERSION_MAP.keys())
        raise ValueError(f"Version {version} is not supported. Supported versions are: {supported_versions}")


class OpenXpdUUID(str):
    """
    An open xpd UUID format for IDs of openEPD documents.

    See https://github.com/cchangelabs/open-xpd-uuid-lib for details.
    """

    def _validate_id(cls, v: str | None) -> str | None:
        if v is None:
            return v

        try:
            open_xpd_uuid.validate(open_xpd_uuid.sanitize(str(v)))
            return v
        except open_xpd_uuid.GuidValidationError as e:
            raise ValueError("Invalid format") from e

    @classmethod
    def __get_validators__(cls):
        yield cls._validate_id

    @classmethod
    def __modify_schema__(cls, field_schema):
        field_schema.update(
            example="XC300001",
        )



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from openepd.compat.pydantic import pyd
from openepd.model.specs.base import BaseOpenEpdHierarchicalSpec
from openepd.model.specs.enums import ElevatorsBuildingRise, ElevatorsUsageIntensity
from openepd.model.validation.quantity import CapacityPerHourStr, LengthMStr, MassKgStr, SpeedStr


class EscalatorsV1(BaseOpenEpdHierarchicalSpec):
    """
    Escalators category performance specification.

    A moving staircase consisting of a circulating belt of steps driven by a motor,
    which conveys people between the floors of a building.
    """

    vertical_rise: LengthMStr | None = pyd.Field(
        default=None,
        description="The vertical distance between the top and bottom landings of an escalator",
        example="1 m",
    )
    speed: SpeedStr | None = pyd.Field(default=None, description="Reference speed of the escalator", example="1 m / s")
    step_width: LengthMStr | None = pyd.Field(default=None, description="Width of the escalator steps", example="1 m")
    max_capacity: CapacityPerHourStr | None = pyd.Field(
        default=None, description="Max capacity of escalator in persons per hour", example="1 hour^-1"
    )
    indoor: bool | None = pyd.Field(
        default=None, description="Escalator can be used for indoor applications", example=True
    )
    outdoor: bool | None = pyd.Field(
        default=None, description="Escalator can be used for outdoor applications", example=True
    )

    _EXT_VERSION = "1.1"


class ElevatorsV1(BaseOpenEpdHierarchicalSpec):
    """Car that moves in a vertical shaft to carry passengers or freight between the levels of a multistory building."""

    _EXT_VERSION = "1.0"

    # Own fields:
    usage_intensity: list[ElevatorsUsageIntensity] | None = pyd.Field(
        default=None, description="", example=["Very low"]
    )
    travel_length: LengthMStr | None = pyd.Field(default=None, description="", example="1 m")
    rated_load: MassKgStr | None = pyd.Field(default=None, description="", example="1 kg")
    rated_speed: SpeedStr | None = pyd.Field(default=None, description="", example="1 m / s")
    building_rise: ElevatorsBuildingRise | None = pyd.Field(default=None, description="", example="Low-rise")


class ConveyingEquipmentV1(BaseOpenEpdHierarchicalSpec):
    """Conveying Equipment."""

    _EXT_VERSION = "1.1"

    # Nested specs:
    Elevators: ElevatorsV1 | None = None
    Escalators: EscalatorsV1 | None = None



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
__all__ = ("OpenEpdApiClientSync",)

from requests.auth import AuthBase

from openepd.api.average_dataset.generic_estimate_sync_api import GenericEstimateApi
from openepd.api.average_dataset.industry_epd_sync_api import IndustryEpdApi
from openepd.api.base_sync_client import SyncHttpClient, TokenAuth
from openepd.api.category.sync_api import CategoryApi
from openepd.api.epd.sync_api import EpdApi
from openepd.api.pcr.sync_api import PcrApi


class OpenEpdApiClientSync:
    """Synchronous API client for OpenEPD."""

    def __init__(self, base_url: str, auth_token: str | None, **kwargs) -> None:
        """
        Construct an API client.

        :param base_url: base URL of the API
        :param auth_token: authentication token
        :param kwargs: additional arguments to pass to the HTTP client. See SyncHttpClient constructor for details.
        """
        super().__init__()
        auth: AuthBase | None = TokenAuth(auth_token) if auth_token is not None else None
        self._http_client = SyncHttpClient(base_url, auth=auth, **kwargs)
        self.__epd_api: EpdApi | None = None
        self.__pcr_api: PcrApi | None = None
        self.__category_api: CategoryApi | None = None
        self.__generic_estimate_api: GenericEstimateApi | None = None
        self.__industry_epd_api: IndustryEpdApi | None = None

    @property
    def epds(self) -> EpdApi:
        """Get the EPD API."""
        if self.__epd_api is None:
            self.__epd_api = EpdApi(self._http_client)
        return self.__epd_api

    @property
    def pcrs(self) -> PcrApi:
        """Get the PCR API."""
        if self.__pcr_api is None:
            self.__pcr_api = PcrApi(self._http_client)
        return self.__pcr_api

    @property
    def categories(self) -> CategoryApi:
        """Get the Category API."""
        if self.__category_api is None:
            self.__category_api = CategoryApi(self._http_client)
        return self.__category_api

    @property
    def industry_epds(self) -> IndustryEpdApi:
        """Get the Category API."""
        if self.__industry_epd_api is None:
            self.__industry_epd_api = IndustryEpdApi(self._http_client)
        return self.__industry_epd_api

    @property
    def generic_estimates(self) -> GenericEstimateApi:
        """Get the GE API."""
        if self.__generic_estimate_api is None:
            self.__generic_estimate_api = GenericEstimateApi(self._http_client)
        return self.__generic_estimate_api



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from datetime import datetime
from enum import StrEnum

from openepd.compat.pydantic import pyd
from openepd.model.base import BaseOpenEpdSchema


class BundleManifestAssetsStats(BaseOpenEpdSchema):
    """The statistics of assets in a bundle."""

    total_count: int = 0
    """The total number of assets."""

    total_size: int = 0
    """The total size of assets in bytes."""

    count_by_type: dict[str, int] = pyd.Field(default_factory=dict)
    """The number of assets by type."""


class AssetType(StrEnum):
    """The type of asset."""

    Epd = "epd"
    Pcr = "pcr"
    Org = "org"
    Blob = "blob"


class RelType(StrEnum):
    """The type of relationship between assets."""

    Translation = "translation"
    """A translation of the asset."""
    Pdf = "repr.pdf"
    """A PDF representation of the asset."""
    Ilcd = "repr.ilcd"
    """An ILCD representation of the asset."""


class BundleManifest(BaseOpenEpdSchema):
    """The manifest of a bundle."""

    format: str = "openEPD Bundle/1.0"
    """The format of the bundle."""
    generator: str
    """The generator of the bundle."""
    assets: BundleManifestAssetsStats = pyd.Field(default_factory=BundleManifestAssetsStats)
    comment: str | None = pyd.Field(default=None)
    created_at: datetime = pyd.Field(default_factory=datetime.utcnow)
    """The date and time when the bundle was generated."""


class AssetInfo(BaseOpenEpdSchema):
    """A table of contents item."""

    ref: str
    """The ID of the asset."""
    name: str | None = None
    """The name of the asset."""
    type: AssetType
    """The type of the asset."""
    lang: str | None
    """The language of the asset."""
    rel_type: str | None
    rel_asset: str | None
    comment: str | None = pyd.Field(default=None)
    content_type: str | None = pyd.Field(default=None)
    size: int | None = pyd.Field(default=None)
    custom_type: str | None = pyd.Field(default=None)
    custom_data: str | None = pyd.Field(default=None)



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
__all__ = (
    "GMURangeV1",
    "AutoclavedAeratedConcreteRangeV1",
    "BrickRangeV1",
    "MasonryRangeV1",
)

# NB! This is a generated code. Do not edit it manually. Please see src/openepd/model/specs/README.md


from openepd.compat.pydantic import pyd
from openepd.model.specs.base import BaseOpenEpdHierarchicalSpec
from openepd.model.validation.quantity import AmountRangePressureMpa, AmountRangeThermalConductivity


class GMURangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Glass masonry unit.

    Range version.
    """

    _EXT_VERSION = "1.0"


class AutoclavedAeratedConcreteRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    A lightweight, precast, foamed concrete masonry building material.

    Range version.
    """

    _EXT_VERSION = "1.0"

    strength_28d: AmountRangePressureMpa | None = pyd.Field(default=None, description="")
    thermal_conductivity: AmountRangeThermalConductivity | None = pyd.Field(default=None, description="")
    white: bool | None = pyd.Field(default=None, description="")


class BrickRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Solid masonry units made from clay or shale.

    Range version.
    """

    _EXT_VERSION = "1.0"

    building: bool | None = pyd.Field(default=None, description="")
    facing: bool | None = pyd.Field(default=None, description="")
    floor: bool | None = pyd.Field(default=None, description="")
    pedestrian: bool | None = pyd.Field(default=None, description="")
    paving: bool | None = pyd.Field(default=None, description="")
    other: bool | None = pyd.Field(default=None, description="")
    chemical_resistant: bool | None = pyd.Field(default=None, description="")
    glazed: bool | None = pyd.Field(default=None, description="")
    tiles: bool | None = pyd.Field(default=None, description="")


class MasonryRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Structural and/or enclosure system based on individual rigid units stacked and bound together with mortar.

    Range version.
    """

    _EXT_VERSION = "1.0"

    white_cement: bool | None = pyd.Field(default=None, description="")
    GMU: GMURangeV1 | None = None
    AutoclavedAeratedConcrete: AutoclavedAeratedConcreteRangeV1 | None = None
    Brick: BrickRangeV1 | None = None



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from openepd.compat.pydantic import pyd
from openepd.model.specs.base import BaseOpenEpdHierarchicalSpec
from openepd.model.specs.enums import (
    CableTraysMaterial,
    CablingCategory,
    CablingFireRating,
    CablingJacketMaterial,
    FloorBoxCoverMaterial,
    FloorBoxFloorMaterial,
    FloorBoxMaterial,
    PduTechnology,
    RacewaysMaterial,
    RackType,
)
from openepd.model.specs.singular.mixins.conduit_mixin import ConduitMixin
from openepd.model.validation.quantity import (
    ElectricalCurrentStr,
    LengthMmStr,
    MassKgStr,
    validate_quantity_unit_factory,
)


class PDUV1(BaseOpenEpdHierarchicalSpec):
    """Devices with multiple outputs designed to distribute power, often in a data center cabinet."""

    _EXT_VERSION = "1.0"

    # Own fields:
    amperage: ElectricalCurrentStr | None = pyd.Field(default=None, description="", example="1 A")
    outlet_level_metering: bool | None = pyd.Field(default=None, description="", example=True)
    outlet_level_switching: bool | None = pyd.Field(default=None, description="", example=True)
    pdu_technology: PduTechnology | None = pyd.Field(default=None, description="", example="Basic")
    pdu_outlets: int | None = pyd.Field(default=None, description="", example=3, le=200)

    _amperage_is_quantity_validator = pyd.validator("amperage", allow_reuse=True)(validate_quantity_unit_factory("A"))


class CabinetsRacksAndEnclosuresV1(BaseOpenEpdHierarchicalSpec):
    """Physical support upon which network equipment is mounted. Includes cabinets, racks, frames, and enclosures."""

    _EXT_VERSION = "1.0"

    # Own fields:
    static_load: MassKgStr | None = pyd.Field(default=None, description="", example="1 kg")
    total_racking_units: int | None = pyd.Field(default=None, description="", example=3)
    rack_type: RackType | None = pyd.Field(default=None, description="", example="Cabinet")


class DataCablingV1(BaseOpenEpdHierarchicalSpec):
    """Telecommunications cabling for buildings."""

    _EXT_VERSION = "1.0"

    # Own fields:
    outdoor: bool | None = pyd.Field(default=None, description="", example=True)
    cabling_category: CablingCategory | None = pyd.Field(default=None, description="", example="Cat7")
    fire_rating: CablingFireRating | None = pyd.Field(default=None, description="", example="CMP")
    jacket_material: CablingJacketMaterial | None = pyd.Field(default=None, description="", example="PVC")
    shielded: bool | None = pyd.Field(
        default=None, description="Foil or similar electromagnetic shielding", example=True
    )
    armored: bool | None = pyd.Field(default=None, description="Steel or similar physical armor jacket", example=True)
    rohs: bool | None = pyd.Field(default=None, description="Certified ROHS Compliant", example=True)
    reach: bool | None = pyd.Field(default=None, description="Certified REACH compliant", example=True)
    zwtl: bool | None = pyd.Field(default=None, description="Certified ZWTL compliant", example=True)
    connectorized: bool | None = pyd.Field(
        default=None,
        description="This cable is shipped as a specific length with integrated connectors. Impacts include the "
        "connectors for the specific cable length. Connectors add impact similar to 0.1-0.5 additional "
        "meters of cable",
        example=True,
    )
    thin_ethernet: bool | None = pyd.Field(
        default=None,
        description="At least part of this cable has a reduced outer diameter and thinner wires. Thin ethernet cables "
        "have handling advantages and tend to have a reduced impact, but also reduced channel length. "
        "See TIA 568.2-D Annex G.",
        example=True,
    )


class FloorBoxesAndAccessoriesV1(BaseOpenEpdHierarchicalSpec):
    """
    Electrical boxes that are installed in the floor.

    Used to provide power and/or data connections to devices in a room or space.
    """

    _EXT_VERSION = "1.0"

    # Own fields:
    painted: bool | None = pyd.Field(default=None, description="", example=True)
    fire_classified: bool | None = pyd.Field(
        default=None, description="Includes hardware to maintain fire rating of the floor", example=True
    )
    outdoor: bool | None = pyd.Field(default=None, description="Floor boxes installed in the ground", example=True)
    raised: bool | None = pyd.Field(default=None, description="Used in raised or computer style flooring", example=True)
    poke_through: bool | None = pyd.Field(
        default=None,
        description="Used primarily in retrofit or renovation and will maintain fire rating of the floor",
        example=True,
    )
    cover: bool | None = pyd.Field(
        default=None, description="Floor box cover or lid for use with a separate floor box", example=True
    )
    outlets: int | None = pyd.Field(
        default=None,
        description="Number of outlet ports from floor box, including power, data, video, and other connections",
        example=3,
        le=16,
    )
    material: FloorBoxMaterial | None = pyd.Field(default=None, description="", example="Metallic Box")
    cover_material: FloorBoxCoverMaterial | None = pyd.Field(default=None, description="", example="Brass")
    floor_material: FloorBoxFloorMaterial | None = pyd.Field(default=None, description="", example="Concrete")


class NetworkingCableTraysV1(BaseOpenEpdHierarchicalSpec):
    """
    Cable trays.

    Mechanical support systems that provide a rigid structural system for cables used for communication and
    power distribution.
    """

    _EXT_VERSION = "1.0"

    # Own fields:
    height: LengthMmStr | None = pyd.Field(default=None, description="", example="100 mm")
    width: LengthMmStr | None = pyd.Field(default=None, description="", example="100 mm")
    depth: LengthMmStr | None = pyd.Field(default=None, description="Depth of enclosure system", example="1 m")
    static_load: MassKgStr | None = pyd.Field(default=None, description="Mass that the unit can hold", example="1 kg")
    ventilated: bool | None = pyd.Field(
        default=None, description="At least 40% of the tray base is open to air flow", example=True
    )
    material: CableTraysMaterial | None = pyd.Field(default=None, description="", example="Stainless Steel")


class NetworkingRacewaysV1(BaseOpenEpdHierarchicalSpec):
    """
    Surface-mounted systems along the perimeter of walls to route, conceal, and protect cables.

    Often called trunking.
    """

    _EXT_VERSION = "1.0"

    # Own fields:
    width: LengthMmStr | None = pyd.Field(default=None, description="", example="100 mm")
    depth: LengthMmStr | None = pyd.Field(default=None, description="Depth of enclosure system", example="100 mm")
    painted: bool | None = pyd.Field(default=None, description="", example=True)
    divided: bool | None = pyd.Field(
        default=None,
        description="Dual service raceway for high and low voltage data and power applications",
        example=True,
    )
    raceways_material: RacewaysMaterial | None = pyd.Field(default=None, description="", example="Aluminum")


class CommunicationsConduitV1(BaseOpenEpdHierarchicalSpec, ConduitMixin):
    """Tubing used to protect and route communications wiring in a building or structure."""

    _EXT_VERSION = "1.1"


class NetworkInfrastructureV1(BaseOpenEpdHierarchicalSpec):
    """General category for network infrastructure products for data centers and commercial and residential buildings."""

    _EXT_VERSION = "1.1"

    # Nested specs:
    PDU: PDUV1 | None = None
    CabinetsRacksAndEnclosures: CabinetsRacksAndEnclosuresV1 | None = None
    DataCabling: DataCablingV1 | None = None
    FloorBoxesAndAccessories: FloorBoxesAndAccessoriesV1 | None = None
    NetworkingCableTrays: NetworkingCableTraysV1 | None = None
    NetworkingRaceways: NetworkingRacewaysV1 | None = None
    CommunicationsConduit: CommunicationsConduitV1 | None = None



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
__all__ = (
    "AcTransformersRangeV1",
    "ElectricalInsulatorsRangeV1",
    "ElectricalSubstationsRangeV1",
    "ElectricalSwitchgearRangeV1",
    "PowerCablingRangeV1",
    "ElectricalTransmissionAndDistributionEquipmentRangeV1",
)

# NB! This is a generated code. Do not edit it manually. Please see src/openepd/model/specs/README.md


from openepd.model.specs.base import BaseOpenEpdHierarchicalSpec


class AcTransformersRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Equipment for transforming between higher and lower voltage AC power.

    Range version.
    """

    _EXT_VERSION = "1.0"


class ElectricalInsulatorsRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Passive electrical isolation.

    Range version.
    """

    _EXT_VERSION = "1.0"


class ElectricalSubstationsRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Electrical substations performance specification.

    Range version.
    """

    _EXT_VERSION = "1.0"


class ElectricalSwitchgearRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Equipment for interrupting and controlling high-power electrical flows.

    Used for protection, isolation, or control of electrical equipment.

    Range version.
    """

    _EXT_VERSION = "1.0"


class PowerCablingRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    High-voltage electrical cabling.

    Range version.
    """

    _EXT_VERSION = "1.0"


class ElectricalTransmissionAndDistributionEquipmentRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Electrical Transmission & Distribution Equipment.

    Range version.
    """

    _EXT_VERSION = "1.0"

    AcTransformers: AcTransformersRangeV1 | None = None
    ElectricalInsulators: ElectricalInsulatorsRangeV1 | None = None
    ElectricalSubstations: ElectricalSubstationsRangeV1 | None = None
    ElectricalSwitchgear: ElectricalSwitchgearRangeV1 | None = None
    PowerCabling: PowerCablingRangeV1 | None = None



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
__all__ = (
    "CementRangeV1",
    "MasonryCementRangeV1",
    "SupplementaryCementitiousMaterialsRangeV1",
    "AccessFlooringPedestalsRangeV1",
    "CarpetBackingRangeV1",
    "CarpetFiberRangeV1",
    "CementitiousMaterialsRangeV1",
    "ConcreteAdmixturesRangeV1",
    "TextilesRangeV1",
    "ManufacturingInputsRangeV1",
)

# NB! This is a generated code. Do not edit it manually. Please see src/openepd/model/specs/README.md


from openepd.compat.pydantic import pyd
from openepd.model.common import RangeFloat
from openepd.model.specs.base import BaseOpenEpdHierarchicalSpec
from openepd.model.specs.concrete import Cementitious
from openepd.model.specs.enums import (
    AdmixtureEffects,
    CarpetYarnType,
    CementAstmType,
    CementC1157,
    CementCsaA3001,
    CementEn197_1,
    CementScm,
    MasonryCementAstmC91Type,
    TextilesFabricType,
)


class CementRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Cements.

    Includes Portland and blended cements, that can serve as the primary binder in a concrete mix.

    Range version.
    """

    _EXT_VERSION = "1.0"

    cementitious: Cementitious | None = pyd.Field(default=None, description="")
    white_cement: bool | None = pyd.Field(default=None, description="")
    astm_type: list[CementAstmType] | None = pyd.Field(default=None, description="")
    c1157: list[CementC1157] | None = pyd.Field(default=None, description="")
    csa_a3001: list[CementCsaA3001] | None = pyd.Field(default=None, description="")
    en197_1: list[CementEn197_1] | None = pyd.Field(default=None, description="")
    oil_well_cement: bool | None = pyd.Field(default=None, description="")


class MasonryCementRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    A cementitious product typically composed of Portland cement and hydrated lime.

    Masonry cement is combined with sand to make mortar.

    Range version.
    """

    _EXT_VERSION = "1.0"

    astm_c91_type: list[MasonryCementAstmC91Type] | None = pyd.Field(default=None, description="")


class SupplementaryCementitiousMaterialsRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Cementitious materials that are not effective binders when used on their own.

    Range version.
    """

    _EXT_VERSION = "1.0"

    cement_scm: list[CementScm] | None = pyd.Field(default=None, description="")


class AccessFlooringPedestalsRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Part of an access floor system.

    Pedestals are laid out on top of a floor slab and support access floor panels, creating a void space between the
    floor slab and finish floor.

    Range version.
    """

    _EXT_VERSION = "1.0"


class CarpetBackingRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Fabric backing holding a carpet together.

    Range version.
    """

    _EXT_VERSION = "1.0"


class CarpetFiberRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Fiber yarn used in the manufacture of carpet, typically nylon or wool.

    Range version.
    """

    _EXT_VERSION = "1.0"

    yarn_material: list[CarpetYarnType] | None = pyd.Field(default=None, description="")
    yarn_recycled_content: RangeFloat | None = pyd.Field(default=None, description="")


class CementitiousMaterialsRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Cementitious materials performance specification.

    Range version.
    """

    _EXT_VERSION = "1.0"

    Cement: CementRangeV1 | None = None
    MasonryCement: MasonryCementRangeV1 | None = None
    SupplementaryCementitiousMaterials: SupplementaryCementitiousMaterialsRangeV1 | None = None


class ConcreteAdmixturesRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Concrete admixtures.

    Chemical additives that are added to fresh concrete immediately before or during mixing. Admixtures have distinct
    functions and are categorized as: air-entraining, water-reducing, retarding, accelerating, and plasticizers
    (i.e., superplasticizers).

    Range version.
    """

    _EXT_VERSION = "1.0"

    effects: list[AdmixtureEffects] | None = pyd.Field(default=None, description="")


class TextilesRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Textiles for use in manufacturing end products.

    Range version.
    """

    _EXT_VERSION = "1.0"

    fabric_type: list[TextilesFabricType] | None = pyd.Field(default=None, description="")


class ManufacturingInputsRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Manufacturing inputs.

    Broad category for collecting materials primarily used as manufacturing inputs, rather than directly used in
    a construction.

    Range version.
    """

    _EXT_VERSION = "1.0"

    AccessFlooringPedestals: AccessFlooringPedestalsRangeV1 | None = None
    CarpetBacking: CarpetBackingRangeV1 | None = None
    CarpetFiber: CarpetFiberRangeV1 | None = None
    CementitiousMaterials: CementitiousMaterialsRangeV1 | None = None
    ConcreteAdmixtures: ConcreteAdmixturesRangeV1 | None = None
    Textiles: TextilesRangeV1 | None = None



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from typing import Any, Callable

from openepd.model.common import EnumGroupingAware


def exclusive_groups_validator_factory(enum_type: type[EnumGroupingAware]) -> Callable[[type, Any], Any]:
    """
    Create an exclusive groups validator.

    If we have a certain enum TheEnum, and a field of list[TheEnum], where list can contain only one value of each of
    the groups, this validator should be used. For example, ACI exposure classes for concrete specify various
    parameters such as water resistance, chemical resistance, etc., and the list of classes can have 0 or 1 from each
    group.

    :param enum_type:Enum type which supports groupings.
    :return:value, or raises ValueError if not allowed combination is given.
    """

    def enum_exclusive_grouping_validator(cls, value: list | None) -> list | None:
        for grouping in enum_type.get_groupings():
            matching_from_group = [v for v in (value or []) if v in grouping]
            if len(matching_from_group) > 1:
                raise ValueError(f"Values {', '.join(matching_from_group)} are not allowed together.")

        return value

    return enum_exclusive_grouping_validator



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
__all__ = (
    "BlanketFacingRangeV1",
    "DoorsHardwareRangeV1",
    "FlooringAccessoriesRangeV1",
    "MortarRangeV1",
    "TileGroutRangeV1",
    "AccessoriesRangeV1",
)

# NB! This is a generated code. Do not edit it manually. Please see src/openepd/model/specs/README.md


from openepd.model.specs.base import BaseOpenEpdHierarchicalSpec


class BlanketFacingRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Facing materials for insulation products.

    Such as kraft, white vinyl sheeting, or aluminum foil, which can serve as air barrier, vapor barrier, radiant
    barrier, or flame resistive layer.

    Range version.
    """

    _EXT_VERSION = "1.0"


class DoorsHardwareRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Door hardware, including automatic and security door hardware.

    Range version.
    """

    _EXT_VERSION = "1.0"


class FlooringAccessoriesRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Products used in flooring, other than the actual flooring product itself.

    Range version.
    """

    _EXT_VERSION = "1.0"


class MortarRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Cementitious paste used to bind building blocks such as stones, bricks, and concrete masonry.

    Range version.
    """

    _EXT_VERSION = "1.0"


class TileGroutRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Water-cement-sand mixture for laying ceramic tile.

    Range version.
    """

    _EXT_VERSION = "1.0"


class AccessoriesRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Materials that are used alongside other materials.

    Range version.
    """

    _EXT_VERSION = "1.0"

    BlanketFacing: BlanketFacingRangeV1 | None = None
    DoorsHardware: DoorsHardwareRangeV1 | None = None
    FlooringAccessories: FlooringAccessoriesRangeV1 | None = None
    Mortar: MortarRangeV1 | None = None
    TileGrout: TileGroutRangeV1 | None = None



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
import dataclasses
from types import UnionType
from typing import Any

from openepd.compat.pydantic import pyd
from openepd.model.base import BaseOpenEpdSchema, Version
from openepd.model.validation.common import validate_version_compatibility, validate_version_format
from openepd.model.validation.quantity import ExternalValidationConfig, QuantityValidator
from openepd.model.versioning import WithExtVersionMixin


class BaseOpenEpdSpec(BaseOpenEpdSchema):
    """Base class for all OpenEPD specs."""

    class Config:
        use_enum_values = False  # we need to store enums as strings and not values


class BaseOpenEpdHierarchicalSpec(BaseOpenEpdSpec, WithExtVersionMixin):
    """Base class for new specs (hierarchical, versioned)."""

    def __init__(self, **data: Any) -> None:
        # ensure that all the concrete spec objects fail on creations if they dont have _EXT_VERSION declared to
        # something meaningful
        if not hasattr(self, "_EXT_VERSION") or self._EXT_VERSION is None:
            raise ValueError(f"Class {self.__class__} must declare an extension version")
        Version.parse_version(self._EXT_VERSION)  # validate format correctness
        super().__init__(**{"ext_version": self._EXT_VERSION, **data})

    _version_format_validator = pyd.validator("ext_version", allow_reuse=True, check_fields=False)(
        validate_version_format
    )
    _version_major_match_validator = pyd.validator("ext_version", allow_reuse=True, check_fields=False)(
        validate_version_compatibility("_EXT_VERSION")
    )


def setup_external_validators(quantity_validator: QuantityValidator):
    """Set the implementation unit validator for specs."""
    ExternalValidationConfig.QUANTITY_VALIDATOR = quantity_validator


@dataclasses.dataclass(kw_only=True)
class CodegenSpec:
    """
    Specification for codegen when generating RangeSpecs from normal specs.

    See openepd.mode.specs.README.md for details.
    """

    exclude_from_codegen: bool = False
    override_type: type | UnionType



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from openepd.model.specs.base import BaseOpenEpdHierarchicalSpec


class AuxiliariesV1(BaseOpenEpdHierarchicalSpec):
    """Auxiliaries performance specification."""

    _EXT_VERSION = "1.0"


class CleaningProductsV1(BaseOpenEpdHierarchicalSpec):
    """Cleaning and disinfecting solutions."""

    _EXT_VERSION = "1.0"


class ClothingV1(BaseOpenEpdHierarchicalSpec):
    """Clothing."""

    _EXT_VERSION = "1.0"


class FoodBeverageV1(BaseOpenEpdHierarchicalSpec):
    """Food Beverage and Tobacco Products."""

    _EXT_VERSION = "1.0"


class TransportationInfrastructureV1(BaseOpenEpdHierarchicalSpec):
    """A broad category for unclassified materials focused on transportation infrastructure."""

    _EXT_VERSION = "1.0"

    # Nested specs:
    Auxiliaries: AuxiliariesV1 | None = None


class UnsupportedV1(BaseOpenEpdHierarchicalSpec):
    """
    A generic category for EPDs/Materials and categories that are explicitly not yet supported by EC3.

    Assume that any data in this subcategory is unreliable.
    """

    _EXT_VERSION = "1.0"

    # Nested specs:
    CleaningProducts: CleaningProductsV1 | None = None
    Clothing: ClothingV1 | None = None
    FoodBeverage: FoodBeverageV1 | None = None


class CopperV1(BaseOpenEpdHierarchicalSpec):
    """Products made of copper."""

    _EXT_VERSION = "1.0"


class EarthworkV1(BaseOpenEpdHierarchicalSpec):
    """Earthwork, including excavation, shoring, piles, etc."""

    _EXT_VERSION = "1.0"


class ExteriorSunControlDevicesV1(BaseOpenEpdHierarchicalSpec):
    """Sun control devices help to manage solar heat gain by redirecting sunlight."""

    _EXT_VERSION = "1.0"


class GypsumFinishingCompoundsV1(BaseOpenEpdHierarchicalSpec):
    """Plasters and the like for finishing Gypsum Sheet and Board."""

    _EXT_VERSION = "1.0"


class ProfilesV1(BaseOpenEpdHierarchicalSpec):
    """Metal or polymer profiles used for producing or installing of Windows, Doors, Frames and Cladding."""

    _EXT_VERSION = "1.0"


class UnknownV1(BaseOpenEpdHierarchicalSpec):
    """Materials with unknown category. Assume that any data in this subcategory is unreliable."""

    _EXT_VERSION = "1.0"


class ZincV1(BaseOpenEpdHierarchicalSpec):
    """Products made of zinc."""

    _EXT_VERSION = "1.0"


class OtherMaterialsV1(BaseOpenEpdHierarchicalSpec):
    """Broad category of materials not yet classified."""

    _EXT_VERSION = "1.0"

    # Nested specs:
    TransportationInfrastructure: TransportationInfrastructureV1 | None = None
    Unsupported: UnsupportedV1 | None = None
    Copper: CopperV1 | None = None
    Earthwork: EarthworkV1 | None = None
    ExteriorSunControlDevices: ExteriorSunControlDevicesV1 | None = None
    GypsumFinishingCompounds: GypsumFinishingCompoundsV1 | None = None
    Profiles: ProfilesV1 | None = None
    Unknown: UnknownV1 | None = None
    Zinc: ZincV1 | None = None



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from openepd.model.base import BaseDocumentFactory, OpenEpdDoctypes, RootDocument
from openepd.model.epd import EpdFactory
from openepd.model.generic_estimate import GenericEstimateFactory
from openepd.model.industry_epd import IndustryEpdFactory


class DocumentFactory:
    """A factory for creating documents regardless of the type."""

    DOCTYPE_TO_FACTORY: dict[OpenEpdDoctypes, type[BaseDocumentFactory]] = {
        OpenEpdDoctypes.Epd: EpdFactory,
        OpenEpdDoctypes.GenericEstimate: GenericEstimateFactory,
        OpenEpdDoctypes.IndustryEpd: IndustryEpdFactory,
    }

    @classmethod
    def get_factory(cls, doctype: OpenEpdDoctypes | None) -> type[BaseDocumentFactory]:
        """
        Get a document factory by given doctype.

        :param doctype: doctype
        :return document factory
        :raise ValueError if doctype not supported or not found.
        """
        if doctype is None:
            raise ValueError("The document type is not specified.")
        factory = cls.DOCTYPE_TO_FACTORY.get(doctype)
        if factory is None:
            raise ValueError(
                f"The document of type `{doctype}` is not supported. Supported documents are: "
                + ", ".join(cls.DOCTYPE_TO_FACTORY.keys())
            )
        return factory

    @classmethod
    def from_dict(cls, data: dict) -> RootDocument:
        """
        Create a document from the dictionary.

        Type of the document will be recognized from the `doctype` field.
        :raise ValueError: if the document type is not specified or not supported.
        """
        doctype = data.get("doctype")
        if doctype is None or not isinstance(doctype, (str, OpenEpdDoctypes)):
            raise ValueError(
                f"The document type is not specified or not supported. "
                f"Please specify it in `doctype` field. Supported are: {','.join(cls.DOCTYPE_TO_FACTORY)}"
            )

        factory = cls.get_factory(OpenEpdDoctypes(doctype))
        return factory.from_dict(data)



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from abc import ABC
from enum import ReprEnum
from typing import ClassVar, NamedTuple

from openepd.compat.pydantic import pyd


class WithExtVersionMixin(ABC, pyd.BaseModel):
    """Mixin for extensions supporting versions: recommended way."""

    _EXT_VERSION: ClassVar[str]
    """Exact version (major, minor) of the spec extension"""

    def __init_subclass__(cls):
        """Set the default value for the ext_version field from _EXT_VERSION class var."""
        super().__init_subclass__()
        if hasattr(cls, "_EXT_VERSION"):
            cls.__fields__["ext_version"].default = cls._EXT_VERSION

    # Note: default is set programmatically in __init_subclass__
    ext_version: str | None = pyd.Field(description="Extension version", example="3.22", default=None)


class Version(NamedTuple):
    """Version of the object or specification."""

    major: int
    minor: int

    @staticmethod
    def parse_version(version: str) -> "Version":
        """Parse the version of extension or the format.

        Version is expected to be major.minor

        :param version: The extension version.
        :return: A tuple of major and minor version numbers.
        """
        splits = version.split(".", 1) if isinstance(version, str) else None
        if len(splits) != 2:
            raise ValueError(f"Invalid version: {version}")
        if not splits[0].isdigit() or not splits[1].isdigit():
            raise ValueError(f"Invalid version: {version}")
        return Version(major=int(splits[0]), minor=int(splits[1]))

    def __str__(self) -> str:
        return self.as_str()

    def __repr__(self) -> str:
        return f"[Version] {self.as_str()}"

    def as_str(self) -> str:
        """Return the version as a string."""
        return f"{self.major}.{self.minor}"


class OpenEpdVersions(Version, ReprEnum):
    """
    Enum of supported openEPD versions.

    When adding a new version - make sure to add a new major version to the list of supported versions.
    When doing non-breaking change - update minor version in a corresponding enum value.
    """

    Version0 = Version(major=0, minor=1)

    @classmethod
    def get_supported_versions(cls) -> list[Version]:
        """Return a list of supported versions."""
        return [x.value for x in cls]

    @classmethod
    def supported_versions_str(cls, major_only: bool = False) -> str:
        """
        Return a comma separated list of the supported versions.

        This is a utility method, might be helpful for logging, building error messages, etc.

        :param major_only: If True, minor component will be replaced with 'x'. E.g. `2.x` instead of `2.1`
        """
        if major_only:
            return ", ".join(f"{x.major}.x" for x in cls.get_supported_versions())
        return ", ".join(str(x) for x in cls.get_supported_versions())

    @classmethod
    def get_most_recent_version(cls, branch: int | None = 0) -> Version:
        """
        Return the most recent version of the openEPD format.

        If branch is specified - returns the most recent version of the specified branch, otherwise returns
        the most recent version among all branches.
        """
        if branch is None:
            highest: Version = max(cls, key=lambda x: x.value.major)  # type: ignore
            if highest:
                return highest
        for x in cls:
            if x.value.major == branch:
                return x.value
        raise ValueError(
            f"No version {branch}.x is not supported. Supported versions are: {', '.join(str(x.value) for x in cls)}"
        )

    @classmethod
    def get_current(cls) -> Version:
        """Return the most recent stable version of the format."""
        return cls.get_most_recent_version()

    def __repr__(self):
        return f"[OpenEpdVesion] {self.name} - {self.value}"



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from typing import Any, ClassVar

from openepd.compat.pydantic import pyd
from openepd.model.specs.base import BaseOpenEpdHierarchicalSpec
from openepd.model.specs.singular.accessories import AccessoriesV1
from openepd.model.specs.singular.aggregates import AggregatesV1
from openepd.model.specs.singular.aluminium import AluminiumV1
from openepd.model.specs.singular.asphalt import AsphaltV1
from openepd.model.specs.singular.bulk_materials import BulkMaterialsV1
from openepd.model.specs.singular.cast_decks_and_underlayment import CastDecksAndUnderlaymentV1
from openepd.model.specs.singular.cladding import CladdingV1
from openepd.model.specs.singular.cmu import CMUV1
from openepd.model.specs.singular.concrete import ConcreteV1
from openepd.model.specs.singular.conveying_equipment import ConveyingEquipmentV1
from openepd.model.specs.singular.deprecated import BaseCompatibilitySpec, get_safely, set_safely
from openepd.model.specs.singular.deprecated.concrete import ConcreteOldSpec
from openepd.model.specs.singular.deprecated.steel import SteelOldSpec
from openepd.model.specs.singular.electrical import ElectricalV1
from openepd.model.specs.singular.electrical_transmission_and_distribution_equipment import (
    ElectricalTransmissionAndDistributionEquipmentV1,
)
from openepd.model.specs.singular.electricity import ElectricityV1
from openepd.model.specs.singular.finishes import FinishesV1
from openepd.model.specs.singular.fire_and_smoke_protection import FireAndSmokeProtectionV1
from openepd.model.specs.singular.furnishings import FurnishingsV1
from openepd.model.specs.singular.grouting import GroutingV1
from openepd.model.specs.singular.manufacturing_inputs import ManufacturingInputsV1
from openepd.model.specs.singular.masonry import MasonryV1
from openepd.model.specs.singular.material_handling import MaterialHandlingV1
from openepd.model.specs.singular.mechanical import MechanicalV1
from openepd.model.specs.singular.mechanical_insulation import MechanicalInsulationV1
from openepd.model.specs.singular.network_infrastructure import NetworkInfrastructureV1
from openepd.model.specs.singular.openings import OpeningsV1
from openepd.model.specs.singular.other_electrical_equipment import OtherElectricalEquipmentV1
from openepd.model.specs.singular.other_materials import OtherMaterialsV1
from openepd.model.specs.singular.plumbing import PlumbingV1
from openepd.model.specs.singular.precast_concrete import PrecastConcreteV1
from openepd.model.specs.singular.sheathing import SheathingV1
from openepd.model.specs.singular.steel import SteelV1
from openepd.model.specs.singular.thermal_moisture_protection import ThermalMoistureProtectionV1
from openepd.model.specs.singular.utility_piping import UtilityPipingV1
from openepd.model.specs.singular.wood import WoodV1
from openepd.model.specs.singular.wood_joists import WoodJoistsV1

__all__ = ("Specs",)


class Specs(BaseOpenEpdHierarchicalSpec):
    """Material specific specs."""

    COMPATIBILITY_SPECS: ClassVar[list[type[BaseCompatibilitySpec]]] = [ConcreteOldSpec, SteelOldSpec]

    _EXT_VERSION = "1.0"

    # Nested specs:
    CMU: CMUV1 | None = None
    Masonry: MasonryV1 | None = None
    Steel: SteelV1 | None = None
    NetworkInfrastructure: NetworkInfrastructureV1 | None = None
    Finishes: FinishesV1 | None = None
    ManufacturingInputs: ManufacturingInputsV1 | None = None
    Accessories: AccessoriesV1 | None = None
    ElectricalTransmissionAndDistributionEquipment: ElectricalTransmissionAndDistributionEquipmentV1 | None = None
    Aggregates: AggregatesV1 | None = None
    ThermalMoistureProtection: ThermalMoistureProtectionV1 | None = None
    Mechanical: MechanicalV1 | None = None
    Aluminium: AluminiumV1 | None = None
    Cladding: CladdingV1 | None = None
    FireAndSmokeProtection: FireAndSmokeProtectionV1 | None = None
    PrecastConcrete: PrecastConcreteV1 | None = None
    Asphalt: AsphaltV1 | None = None
    OtherMaterials: OtherMaterialsV1 | None = None
    Plumbing: PlumbingV1 | None = None
    Electrical: ElectricalV1 | None = None
    UtilityPiping: UtilityPipingV1 | None = None
    BulkMaterials: BulkMaterialsV1 | None = None
    CastDecksAndUnderlayment: CastDecksAndUnderlaymentV1 | None = None
    Concrete: ConcreteV1 | None = None
    Sheathing: SheathingV1 | None = None
    Furnishings: FurnishingsV1 | None = None
    Wood: WoodV1 | None = None
    ConveyingEquipment: ConveyingEquipmentV1 | None = None
    MaterialHandling: MaterialHandlingV1 | None = None
    Openings: OpeningsV1 | None = None
    Electricity: ElectricityV1 | None = None
    Grouting: GroutingV1 | None = None
    MechanicalInsulation: MechanicalInsulationV1 | None = None
    OtherElectricalEquipment: OtherElectricalEquipmentV1 | None = None
    WoodJoists: WoodJoistsV1 | None = None

    # historical backward-compatible specs
    concrete: ConcreteOldSpec | None = None
    steel: SteelOldSpec | None = None

    @pyd.root_validator(pre=True)
    def _ensure_backward_compatibiltiy(cls, values: dict[str, Any]) -> dict[str, Any]:
        """
        Restore the functionality for backward-compatible specs.

        Originally, we used to have 'concrete' and 'steel' specs non-hierarchical and manually maintained. Since we
        introduced hierarchical specs, there is a need to retain the key mapping to the old structure.

        :return: modified values
        """
        for compat_spec in cls.COMPATIBILITY_SPECS:
            if (
                compat_spec.COMPATIBILITY_SPECS_KEY_OLD not in values
                and compat_spec.COMPATIBILITY_SPECS_KEY_NEW not in values
            ):
                continue
            for old_key_spec, new_key_spec in compat_spec.COMPATIBILITY_MAPPING.items():
                has_new_spec, new_value = get_safely(values, new_key_spec)

                # new value is set, we should not override it but should return it in old spec
                if has_new_spec:
                    set_safely(values, old_key_spec, new_value)
                    continue

                has_old_spec, old_value = get_safely(values, old_key_spec)
                # nothing to back
                if not has_old_spec:
                    continue
                set_safely(values, new_key_spec, old_value)
        return values



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
__all__ = (
    "PDURangeV1",
    "CabinetsRacksAndEnclosuresRangeV1",
    "DataCablingRangeV1",
    "FloorBoxesAndAccessoriesRangeV1",
    "NetworkingCableTraysRangeV1",
    "NetworkingRacewaysRangeV1",
    "CommunicationsConduitRangeV1",
    "NetworkInfrastructureRangeV1",
)

# NB! This is a generated code. Do not edit it manually. Please see src/openepd/model/specs/README.md


from openepd.compat.pydantic import pyd
from openepd.model.common import RangeInt
from openepd.model.specs.base import BaseOpenEpdHierarchicalSpec
from openepd.model.specs.enums import (
    CableTraysMaterial,
    CablingCategory,
    CablingFireRating,
    CablingJacketMaterial,
    ConduitMaterial,
    FloorBoxCoverMaterial,
    FloorBoxFloorMaterial,
    FloorBoxMaterial,
    PduTechnology,
    RacewaysMaterial,
    RackType,
)
from openepd.model.validation.quantity import AmountRangeElectricalCurrent, AmountRangeLengthMm, AmountRangeMass


class PDURangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Devices with multiple outputs designed to distribute power, often in a data center cabinet.

    Range version.
    """

    _EXT_VERSION = "1.0"

    amperage: AmountRangeElectricalCurrent | None = pyd.Field(default=None, description="")
    outlet_level_metering: bool | None = pyd.Field(default=None, description="")
    outlet_level_switching: bool | None = pyd.Field(default=None, description="")
    pdu_technology: list[PduTechnology] | None = pyd.Field(default=None, description="")
    pdu_outlets: RangeInt | None = pyd.Field(default=None, description="")


class CabinetsRacksAndEnclosuresRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Physical support upon which network equipment is mounted. Includes cabinets, racks, frames, and enclosures.

    Range version.
    """

    _EXT_VERSION = "1.0"

    static_load: AmountRangeMass | None = pyd.Field(default=None, description="")
    total_racking_units: RangeInt | None = pyd.Field(default=None, description="")
    rack_type: list[RackType] | None = pyd.Field(default=None, description="")


class DataCablingRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Telecommunications cabling for buildings.

    Range version.
    """

    _EXT_VERSION = "1.0"

    outdoor: bool | None = pyd.Field(default=None, description="")
    cabling_category: list[CablingCategory] | None = pyd.Field(default=None, description="")
    fire_rating: list[CablingFireRating] | None = pyd.Field(default=None, description="")
    jacket_material: list[CablingJacketMaterial] | None = pyd.Field(default=None, description="")
    shielded: bool | None = pyd.Field(default=None, description="Foil or similar electromagnetic shielding")
    armored: bool | None = pyd.Field(default=None, description="Steel or similar physical armor jacket")
    rohs: bool | None = pyd.Field(default=None, description="Certified ROHS Compliant")
    reach: bool | None = pyd.Field(default=None, description="Certified REACH compliant")
    zwtl: bool | None = pyd.Field(default=None, description="Certified ZWTL compliant")
    connectorized: bool | None = pyd.Field(
        default=None,
        description="This cable is shipped as a specific length with integrated connectors. Impacts include the connectors for the specific cable length. Connectors add impact similar to 0.1-0.5 additional meters of cable",
    )
    thin_ethernet: bool | None = pyd.Field(
        default=None,
        description="At least part of this cable has a reduced outer diameter and thinner wires. Thin ethernet cables have handling advantages and tend to have a reduced impact, but also reduced channel length. See TIA 568.2-D Annex G.",
    )


class FloorBoxesAndAccessoriesRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Electrical boxes that are installed in the floor.

    Used to provide power and/or data connections to devices in a room or space.

    Range version.
    """

    _EXT_VERSION = "1.0"

    painted: bool | None = pyd.Field(default=None, description="")
    fire_classified: bool | None = pyd.Field(
        default=None, description="Includes hardware to maintain fire rating of the floor"
    )
    outdoor: bool | None = pyd.Field(default=None, description="Floor boxes installed in the ground")
    raised: bool | None = pyd.Field(default=None, description="Used in raised or computer style flooring")
    poke_through: bool | None = pyd.Field(
        default=None, description="Used primarily in retrofit or renovation and will maintain fire rating of the floor"
    )
    cover: bool | None = pyd.Field(default=None, description="Floor box cover or lid for use with a separate floor box")
    outlets: RangeInt | None = pyd.Field(
        default=None,
        description="Number of outlet ports from floor box, including power, data, video, and other connections",
    )
    material: list[FloorBoxMaterial] | None = pyd.Field(default=None, description="")
    cover_material: list[FloorBoxCoverMaterial] | None = pyd.Field(default=None, description="")
    floor_material: list[FloorBoxFloorMaterial] | None = pyd.Field(default=None, description="")


class NetworkingCableTraysRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Cable trays.

    Mechanical support systems that provide a rigid structural system for cables used for communication and
    power distribution.

    Range version.
    """

    _EXT_VERSION = "1.0"

    height: AmountRangeLengthMm | None = pyd.Field(default=None, description="")
    width: AmountRangeLengthMm | None = pyd.Field(default=None, description="")
    depth: AmountRangeLengthMm | None = pyd.Field(default=None, description="Depth of enclosure system")
    static_load: AmountRangeMass | None = pyd.Field(default=None, description="Mass that the unit can hold")
    ventilated: bool | None = pyd.Field(default=None, description="At least 40% of the tray base is open to air flow")
    material: list[CableTraysMaterial] | None = pyd.Field(default=None, description="")


class NetworkingRacewaysRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Surface-mounted systems along the perimeter of walls to route, conceal, and protect cables.

    Often called trunking.

    Range version.
    """

    _EXT_VERSION = "1.0"

    width: AmountRangeLengthMm | None = pyd.Field(default=None, description="")
    depth: AmountRangeLengthMm | None = pyd.Field(default=None, description="Depth of enclosure system")
    painted: bool | None = pyd.Field(default=None, description="")
    divided: bool | None = pyd.Field(
        default=None, description="Dual service raceway for high and low voltage data and power applications"
    )
    raceways_material: list[RacewaysMaterial] | None = pyd.Field(default=None, description="")


class CommunicationsConduitRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Tubing used to protect and route communications wiring in a building or structure.

    Range version.
    """

    _EXT_VERSION = "1.1"

    nominal_diameter: AmountRangeLengthMm | None = None
    outer_diameter: AmountRangeLengthMm | None = None
    inner_diameter: AmountRangeLengthMm | None = None
    wall_thickness: AmountRangeLengthMm | None = None
    material: list[ConduitMaterial] | None = None


class NetworkInfrastructureRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    General category for network infrastructure products for data centers and commercial and residential buildings.

    Range version.
    """

    _EXT_VERSION = "1.1"

    PDU: PDURangeV1 | None = None
    CabinetsRacksAndEnclosures: CabinetsRacksAndEnclosuresRangeV1 | None = None
    DataCabling: DataCablingRangeV1 | None = None
    FloorBoxesAndAccessories: FloorBoxesAndAccessoriesRangeV1 | None = None
    NetworkingCableTrays: NetworkingCableTraysRangeV1 | None = None
    NetworkingRaceways: NetworkingRacewaysRangeV1 | None = None
    CommunicationsConduit: CommunicationsConduitRangeV1 | None = None



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from openepd.compat.pydantic import pyd
from openepd.model.base import BaseOpenEpdSchema
from openepd.model.org import OrgRef
from openepd.model.validation.common import ReferenceStr


class StandardRef(BaseOpenEpdSchema):
    """Reference version (short) of Standard."""

    ref: ReferenceStr | None = pyd.Field(
        default=None,
        example="https://openepd.buildingtransparency.org/api/standards/EN15804",
        description="Reference to this Standard's JSON object",
    )
    short_name: str | None = pyd.Field(description="Short-form of name of standard.  Must be unique. Case-insensitive")


class Standard(StandardRef):
    """A standard, such as EN 15804, ISO 14044, ISO 14024:2018, etc."""

    name: str | None = pyd.Field(description="Full document name.  Must be unique. Case-insensitive", default=None)
    link: pyd.AnyUrl | None = pyd.Field(
        description="Link to the exact standard (including version) referred to", default=None
    )
    issuer: OrgRef | None = pyd.Field(description="Org that issued this standard", default=None)



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
__all__ = (
    "BathtubsRangeV1",
    "FaucetsRangeV1",
    "OtherPlumbingFixturesRangeV1",
    "WaterClosetsRangeV1",
    "FireProtectionPipingRangeV1",
    "FireSprinklersRangeV1",
    "StorageTanksRangeV1",
    "WaterHeatersRangeV1",
    "PlumbingFixturesRangeV1",
    "FireSuppressionRangeV1",
    "PipingRangeV1",
    "PlumbingEquipmentRangeV1",
    "PlumbingRangeV1",
)

# NB! This is a generated code. Do not edit it manually. Please see src/openepd/model/specs/README.md


from openepd.compat.pydantic import pyd
from openepd.model.common import RangeAmount
from openepd.model.specs.base import BaseOpenEpdHierarchicalSpec
from openepd.model.specs.enums import FireProtectionPipingMaterial, PipingAnsiSchedule, PlumbingPipingMaterial
from openepd.model.validation.quantity import AmountRangeLengthMm


class BathtubsRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Bathtubs.

    Range version.
    """

    _EXT_VERSION = "1.0"


class FaucetsRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Faucets.

    Range version.
    """

    _EXT_VERSION = "1.0"


class OtherPlumbingFixturesRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Other plumbing fixtures.

    Range version.
    """

    _EXT_VERSION = "1.0"


class WaterClosetsRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Water Closets.

    Range version.
    """

    _EXT_VERSION = "1.0"


class FireProtectionPipingRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    System of pipes used to supply fire-suppression fluids to homes and/or businesses.

    Range version.
    """

    _EXT_VERSION = "1.0"

    thickness: AmountRangeLengthMm | None = pyd.Field(default=None, description="")
    piping_diameter: AmountRangeLengthMm | None = pyd.Field(default=None, description="")
    mass_per_unit_length: RangeAmount | None = pyd.Field(default=None, description="")
    piping_ansi_schedule: list[PipingAnsiSchedule] | None = pyd.Field(default=None, description="")
    fire_protection_piping_material: list[FireProtectionPipingMaterial] | None = pyd.Field(default=None, description="")


class FireSprinklersRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Fire sprinklers.

    Range version.
    """

    _EXT_VERSION = "1.0"


class StorageTanksRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Storage tanks.

    Range version.
    """

    _EXT_VERSION = "1.0"


class WaterHeatersRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Water heaters.

    Range version.
    """

    _EXT_VERSION = "1.0"


class PlumbingFixturesRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Residential and commercial water closets, urinals, lavatories, sinks, bathtubs, showers, faucets, etc.

    Range version.
    """

    _EXT_VERSION = "1.0"

    Bathtubs: BathtubsRangeV1 | None = None
    Faucets: FaucetsRangeV1 | None = None
    OtherPlumbingFixtures: OtherPlumbingFixturesRangeV1 | None = None
    WaterClosets: WaterClosetsRangeV1 | None = None


class FireSuppressionRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Systems used to extinguish, control, or prevent fires.

    Range version.
    """

    _EXT_VERSION = "1.0"

    FireProtectionPiping: FireProtectionPipingRangeV1 | None = None
    FireSprinklers: FireSprinklersRangeV1 | None = None


class PipingRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Piping.

    System of pipes used to provide water and fuel, remove wastewater, allow venting of gases, or supply
    fire-suppression fluids to homes, businesses, or other facilities.

    Range version.
    """

    _EXT_VERSION = "1.0"

    thickness: AmountRangeLengthMm | None = pyd.Field(default=None, description="")
    piping_diameter: AmountRangeLengthMm | None = pyd.Field(default=None, description="")
    mass_per_unit_length: RangeAmount | None = pyd.Field(default=None, description="")
    piping_ansi_schedule: list[PipingAnsiSchedule] | None = pyd.Field(default=None, description="")
    plumbing_piping_material: list[PlumbingPipingMaterial] | None = pyd.Field(default=None, description="")


class PlumbingEquipmentRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Water softeners, filtration equipment, water heaters, and other plumbing equipment.

    Range version.
    """

    _EXT_VERSION = "1.0"

    StorageTanks: StorageTanksRangeV1 | None = None
    WaterHeaters: WaterHeatersRangeV1 | None = None


class PlumbingRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Residential and commercial plumbing equipment and fixtures.

    Range version.
    """

    _EXT_VERSION = "1.0"

    PlumbingFixtures: PlumbingFixturesRangeV1 | None = None
    FireSuppression: FireSuppressionRangeV1 | None = None
    Piping: PipingRangeV1 | None = None
    PlumbingEquipment: PlumbingEquipmentRangeV1 | None = None



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from typing import Any, ClassVar

from openepd.compat.pydantic import pyd
from openepd.model.base import BaseOpenEpdSchema


class BaseCompatibilitySpec(BaseOpenEpdSchema):
    """
    Base class for compatibility (legacy) specs.

    See Specs model for implementation.
    """

    COMPATIBILITY_SPECS_KEY_OLD: ClassVar[str]
    COMPATIBILITY_SPECS_KEY_NEW: ClassVar[str]
    COMPATIBILITY_MAPPING: ClassVar[dict[str, str]]


def get_safely(d: dict, path: str) -> tuple[bool, Any]:
    """
    Get a value from a mixed object via dotted path.

    Mixed object can be a combination of dicts and pydatnic Models on any hierarchy.
    :param d: source dict/object to search in
    :param path: dotted path in object like specs.Concrete.strength_28d
    :return: tuple (WasFound, Value). First element tells if value was found, second - the value itself.
    """
    path_elements = path.split(".")
    current: Any = d
    for p in path_elements:
        match current:
            case dict():
                if p not in current:
                    return False, None
                current = current.get(p)
            case pyd.BaseModel() as model:
                if not hasattr(current, p) or p not in model.__fields_set__:
                    return False, None
                current = getattr(current, p)
            case _:
                return False, None

    return True, current


def set_safely(d: dict, path: str, value: Any) -> None:
    """
    Safely set an element in a dict.

    Warning: this is asymmetric compared to get_safely, since it sets value in a dict, not in pydantic model object.

    :param d: source dict/object to set value in
    :param path: dotted path in object like specs.Concrete.strength_28d
    :param value: value to set
    :return: None
    """
    path_elements = path.split(".")
    current = d
    for p in path_elements[:-1]:
        if current.get(p) is None:
            current[p] = {}
        current = current[p]

    current[path_elements[-1]] = value



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from enum import StrEnum
from typing import Annotated, Any

from openepd.compat.pydantic import pyd
from openepd.model.base import BaseOpenEpdSchema
from openepd.model.validation.numbers import RatioFloat


class Amount(BaseOpenEpdSchema):
    """A value-and-unit pairing for amounts that do not have an uncertainty."""

    qty: float | None = pyd.Field(description="How much of this in the amount.", ge=0, default=None)
    unit: str | None = pyd.Field(description="Which unit.  SI units are preferred.", example="kg", default=None)

    @pyd.root_validator
    def check_qty_or_unit(cls, values: dict[str, Any]):
        """Ensure that qty or unit is provided."""
        if values.get("qty") is None and values.get("unit") is None:
            raise ValueError("Either qty or unit must be provided.")
        return values

    def to_quantity_str(self):
        """Return a string representation of the amount."""
        return f"{self.qty or ''} {self.unit or 'str'}".strip()


class Distribution(StrEnum):
    """
    Distribution of the measured value.

     * log-normal: Probability distribution of any random parameter whose natural log is normally distributed (the
       PDF is gaussian).
     * normal: Probability distribution of any random parameter whose value is normally distributed around the mean
       (the PDF is gaussian).
     * Continuous uniform probability distribution between minimum value and maximum value and "0" probability beyond
       these.
     * Probability distribution of any random parameter between minimum value and maximum value with the highest
       probability at the average value of minimum plus maximum value. Linear change of probability between minimum,
       maximum and average value.
     * Means Impact is not known, but with >95% certainty the true value is below the declared value.
       So [1e-6,"kgCFC11e",0,"max"] means the ODP was not exactly measured, but it is guaranteed to be below
       1E-6 kg CO2e.  It is acceptable to treat a 'max' distribution a normal or lognormal distribution with variation
       0.1%.  This is conservative, because the 'max' value is usually much greater than the true impact.
    """

    LOG_NORMAL = "log-normal"
    NORMAL = "normal"
    UNIFORM = "uniform"
    TRIANGULAR = "triangular"
    MAX = "max"


class Measurement(BaseOpenEpdSchema):
    """A scientific value with units and uncertainty."""

    mean: float = pyd.Field(description="Mean (expected) value of the measurement")
    unit: str | None = pyd.Field(description="Measurement unit")
    rsd: pyd.PositiveFloat | None = pyd.Field(
        description="Relative standard deviation, i.e. standard_deviation/mean", default=None
    )
    dist: Distribution | None = pyd.Field(
        description="Statistical distribution of the measurement error.", default=None
    )


class IngredientEvidenceTypeEnum(StrEnum):
    """Supported types of evidence for indirect ingredient."""

    PRODUCT_EPD = "Product EPD"


class Ingredient(BaseOpenEpdSchema):
    """
    An ingredient of a product.

    The Ingredients list gives the core data references and quantities. This list is used to document supply-chain
    transparency, such as the EPDs of major components (e.g. cement in concrete, or recycled steel
    in hot-rolled sections).

    Since the exact ingredients may be viewed as a proprietary information by Manufacturers, this schema also allows
    to pass some data about ingredient about explicitly saying what it is. Further, this data can be used to
    calculate the supply chain specificity of the product and uncertainty adjusted factor. For this option, use
    gwp_fraction, citation and evidence_type.
    """

    qty: float | None = pyd.Field(
        description="Number of declared units of this consumed. Negative values indicate an outflow."
    )
    link: pyd.AnyUrl | None = pyd.Field(
        description="Link to this object's OpenEPD declaration. "
        "An OpenIndustryEPD or OpenLCI link is also acceptable.",
        default=None,
    )

    gwp_fraction: RatioFloat | None = pyd.Field(
        default=None,
        description="Fraction of product's A1-A3 GWP this flow represents.  This value, along with the specificity of "
        "the reference, are used to caclulate supply chain specificity.",
    )
    evidence_type: IngredientEvidenceTypeEnum | None = pyd.Field(
        default=None, description="Type of evidence used, which can be used to calculate degree of specificity"
    )
    citation: str | None = pyd.Field(default=None, description="Text citation describing the data source ")

    @pyd.root_validator(skip_on_failure=True)
    def _validate_evidence(cls, values: dict[str, Any]) -> dict[str, Any]:
        # gwp_fraction should be backed by some type of evidence (fraction coming from product EPD etc) to be accounted
        # for in the calculation of uncertainty
        if values.get("gwp_fraction"):
            if not values.get("evidence_type"):
                raise ValueError("evidence_type is required if gwp_fraction is provided")
            if not (values.get("citation") or values.get("link")):
                raise ValueError("link or citation is required if gwp_fraction is provided")

        return values


class LatLng(BaseOpenEpdSchema):
    """A latitude and longitude."""

    lat: float = pyd.Field(description="Latitude", example=47.6062)
    lng: float = pyd.Field(description="Longitude", example=-122.3321)


class Location(BaseOpenEpdSchema):
    """A location on the Earth's surface."""

    pluscode: str | None = pyd.Field(default=None, description="Open Location code of this location")
    latlng: LatLng | None = pyd.Field(default=None, description="Latitude and longitude of this location")
    address: str | None = pyd.Field(default=None, description="Text address, preferably geocoded")
    country: str | None = pyd.Field(default=None, description="2-alpha country code")
    jurisdiction: str | None = pyd.Field(
        default=None, description="Province, State, or similar subdivision below the level of a country"
    )


ATTACHMENT_KNOWN_KEYS: dict[str, str] = {
    "OpenEPD": "A Product EPD in OpenEPD format",
    "openEPD": "Equivalent to OpenEPD",
    "ILCD_EPD": "A Product EPD in ILCD+EPD format.  An underscore is used in place of '+' for compatability purposes.",
    "OpenIndustryEPD": "An Industry EPD in OpenEPD format",
    "openIndustryEPD": "Equivalent to OpenIndustryEPD",
    "LCA_Model": "An underlying LCA model from which one could replicate these EPD results, in LCA Commons 2.0 format",
    "LCA_Report": "An underlying LCA report, in PDF or other document format",
    "lca_software": "A url, optionally with an #anchor tag, pointing to the version of the software used, generally one provided by the software vendor. ",
    "lca_dataset_reference": "A URL link to an lca dataset used in the analysis.  Multiple dataset references can be added by appending a number or string after _reference, e.g. lca_dataset_reference_3 or lca_dataset_reference_steel",
    "EPD": "Environmental Product Declaration, verified by a third party, not in OpenEPD format.",
    "IndustryEPD": "EPD for an industry, sector, or group of companies, not in OpenEPD format.",
    "Datasheet": "A technical data sheet describing the product.",
    "PCR": "Product Category Rules for EPD generation",
    "Contact_Us": "A url to contact.  May be an email (mailto:) link.",
    "URL": "A link to a relevant resource.  No particular format is specified.  Each URL should be uniquely named. ",
    "VOC": "Volatile Organic Compound declaration",
    "HPD": "Material Health Product Declaration",
    "PEF": "Product Environmental Footprint (without 3rd party verification)",
    "MSDS": "Material Safety Data Sheet",
    "Developer": "Link to the website of the group who performed the LCA and/or prepared the EPD.",
}


class AttachmentDict(dict[str, pyd.AnyUrl]):
    """Special form of dict for attachments."""

    @classmethod
    def __modify_schema__(cls, field_schema: dict[str, Any], field: pyd.fields.ModelField | None):
        # This may be generalized later to combine, for example, enum descriptions and field descriptions to provide
        # a better result.
        field_description = field.field_info.description if field else ""
        if field_description:
            field_description = field_description.strip()
            if not field_description.endswith("."):
                field_description += "."
            field_description += " "

        field_schema["description"] = field_description + "Extra properties of string -> URL allowed."
        field_schema["properties"] = {
            k: {"type": "string", "format": "uri", "description": v} for k, v in ATTACHMENT_KNOWN_KEYS.items()
        }
        field_schema["additionalProperties"] = True


class WithAttachmentsMixin(pyd.BaseModel):
    """Mixin for objects that can have attachments."""

    attachments: AttachmentDict = pyd.Field(
        description="Dict of URLs relevant to this entry",
        example={
            "Contact Us": "https://www.c-change-labs.com/en/contact-us/",
            "LinkedIn": "https://www.linkedin.com/company/c-change-labs/",
        },
        default=None,
    )

    def set_attachment(self, name: str, url: str):
        """Set an attachment."""
        if self.attachments is None:
            self.attachments = {}  # type: ignore
        self.attachments[name] = url  # type: ignore

    def set_attachment_if_any(self, name: str, url: str | None):
        """Set an attachment if url is not None."""
        if url is not None:
            self.set_attachment(name, url)


class WithAltIdsMixin(pyd.BaseModel):
    """Mixin for objects that can have alt_ids."""

    alt_ids: dict[Annotated[str, pyd.Field(max_length=200)], str] | None = pyd.Field(
        description="Dict identifiers for this entry.",
        example={
            "oekobau.dat": "bdda4364-451f-4df2-a68b-5912469ee4c9",
        },
        default=None,
    )

    def set_alt_id(self, domain_name: str, value: str):
        """Set an alt_id."""
        if self.alt_ids is None:
            self.alt_ids = {}
        self.alt_ids[domain_name] = value

    def set_alt_id_if_any(self, domain_name: str, value: str | None):
        """Set an alt_id if value is not None."""
        if value is not None:
            self.set_alt_id(domain_name, value)


class OpenEPDUnit(StrEnum):
    """OpenEPD allowed units."""

    kg = "kg"
    m2 = "m2"
    m = "m"
    M2_RSI = "m2 * RSI"
    MJ = "MJ"
    t_km = "t * km"
    MPa = "MPa"
    item = "item"
    W = "W"
    use = "use"
    degree_c = "°C"
    kg_co2 = "kgCO2e"
    hour = "hour"


class RangeBase(BaseOpenEpdSchema):
    """Base class for range types having min and max and order between them."""

    @pyd.root_validator
    def _validate_range_bounds(cls, values: dict[str, Any]) -> dict[str, Any]:
        min_boundary = values.get("min")
        max_boundary = values.get("max")
        if min_boundary is not None and max_boundary is not None and min_boundary > max_boundary:
            raise ValueError("Max should be greater than min")
        return values


class RangeFloat(RangeBase):
    """Structure representing a range of floats."""

    min: float | None = pyd.Field(default=None, example=3.1)
    max: float | None = pyd.Field(default=None, example=5.8)


class RangeInt(RangeBase):
    """Structure representing a range of ints1."""

    min: int | None = pyd.Field(default=None, example=2)
    max: int | None = pyd.Field(default=None, example=3)


class RangeRatioFloat(RangeFloat):
    """Range of ratios (0-1 to 0-10)."""

    min: float | None = pyd.Field(default=None, example=0.2, ge=0, le=1)
    max: float | None = pyd.Field(default=None, example=0.65, ge=0, le=1)


class RangeAmount(RangeFloat):
    """Structure representing a range of quantities."""

    unit: str | None = pyd.Field(default=None, description="Unit of the range.")


class EnumGroupingAware:
    """
    Mixin for enums to support groups.

    With the groups, enum can group its values into more than one groups, so that the validator higher in code can check
    for mutual exclusivity, for example that only one value from the group is permitted at the same time.
    """

    @classmethod
    def get_groupings(cls) -> list[list]:
        """Return logical groupings of the values."""
        return []



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
__all__ = (
    "ColdFormedFramingRangeV1",
    "DeckingSteelRangeV1",
    "SteelSuspensionAssemblyRangeV1",
    "HollowSectionsRangeV1",
    "HotRolledSectionsRangeV1",
    "PlateSteelRangeV1",
    "MetalRailingsRangeV1",
    "MetalStairsRangeV1",
    "MiscMetalFabricationRangeV1",
    "OpenWebMembranesRangeV1",
    "MBQSteelRangeV1",
    "CoilSteelRangeV1",
    "ColdFormedSteelRangeV1",
    "StructuralSteelRangeV1",
    "PrefabricatedSteelAssembliesRangeV1",
    "PostTensioningSteelRangeV1",
    "RebarSteelRangeV1",
    "WireMeshSteelRangeV1",
    "SteelRangeV1",
)

# NB! This is a generated code. Do not edit it manually. Please see src/openepd/model/specs/README.md


from openepd.compat.pydantic import pyd
from openepd.model.common import RangeFloat, RangeRatioFloat
from openepd.model.specs.base import BaseOpenEpdHierarchicalSpec
from openepd.model.specs.enums import SteelComposition, SteelRebarGrade
from openepd.model.specs.singular.steel import SteelMakingRoute
from openepd.model.standard import Standard
from openepd.model.validation.quantity import (
    AmountRangeLengthMm,
    AmountRangePressureMpa,
    AmountRangeThermalConductivity,
    AmountRangeThermalExpansion,
)


class ColdFormedFramingRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Cold Formed Framing performance specification.

    Cold formed steel elements such as studs and framing, typically made from coil or sheet steel and used
    within walls and ceilings.

    Range version.
    """

    _EXT_VERSION = "1.0"


class DeckingSteelRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Corrugated Decking made from cold-formed sheet steel. Often filled with concrete.

    Range version.
    """

    _EXT_VERSION = "1.0"


class SteelSuspensionAssemblyRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Steel suspension assemblies for suspended (e.g. acoustical) ceiling systems.

    Range version.
    """

    _EXT_VERSION = "1.0"


class HollowSectionsRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Hollow cross section steel shape, typically referred to as hollow structural section (HSS).

    Range version.
    """

    _EXT_VERSION = "1.0"

    fabricated: bool | None = None


class HotRolledSectionsRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Hot rolled sections performance specification.

    Steel shapes, such as angles, wide flange beams and I-beams, produced using a high temperature
    mill process.

    Range version.
    """

    _EXT_VERSION = "1.0"

    fabricated: bool | None = None


class PlateSteelRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Plate Steels.

    Flat hot-rolled steel, typically thicker than 'sheet', made by compressing multiple steel
    layers together into one.


    Range version.
    """

    _EXT_VERSION = "1.0"

    fabricated: bool | None = None


class MetalRailingsRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Metal Railings including pipe and tube railings.

    Range version.
    """

    _EXT_VERSION = "1.0"


class MetalStairsRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Metal stairs.

    Includes: metal pan stairs, metal floor plate stairs, grating stairs, fire escapes,
    ladders, and walkways/catwalks/ramps/platforms.

    Range version.
    """

    _EXT_VERSION = "1.0"


class MiscMetalFabricationRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Prefabricated steel assemblies not included in another category.

    Range version.
    """

    _EXT_VERSION = "1.0"


class OpenWebMembranesRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Open web membranes performance specification.

    Lightweight steel truss, typically made of parallel chords and a triangulated web system,
    "proportioned to span between bearing points.

    Range version.
    """

    _EXT_VERSION = "1.0"


class MBQSteelRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Merchant Bar Quality (MBQ) steel.

    Used as feedstock to steel construction products, but also includes rounds, angles, and light structural shapes.

    Range version.
    """

    _EXT_VERSION = "1.0"


class CoilSteelRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Sheet or strip steel, sold in rolls.

    Typically, coil steel is cold-formed into light gauge products.

    Range version.
    """

    _EXT_VERSION = "1.0"


class ColdFormedSteelRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Cold Formed Steel Products.

    All types of cold formed steel products. These products are made from hot-rolled steel coils and
    sheets and are cold formed into products such as studs, decking, panels, and other accessories.

    Range version.
    """

    _EXT_VERSION = "1.0"

    ColdFormedFraming: ColdFormedFramingRangeV1 | None = None
    DeckingSteel: DeckingSteelRangeV1 | None = None
    SteelSuspensionAssembly: SteelSuspensionAssemblyRangeV1 | None = None


class StructuralSteelRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Structural Steel.

    Hot rolled steel shapes, Hollow Sections, pipes, and similar hot-worked structural steels.

    Range version.
    """

    _EXT_VERSION = "1.0"

    modulus_of_elasticity: AmountRangePressureMpa | None = pyd.Field(
        default=None,
        description="Modulus of Elasticity, https://en.wikipedia.org/wiki/Elastic_modulus ",
    )
    thermal_expansion: AmountRangeThermalExpansion | None = pyd.Field(
        default=None,
        description="Thermal Expansion, https://en.wikipedia.org/wiki/Thermal_expansion",
    )
    thermal_conductivity: AmountRangeThermalConductivity | None = pyd.Field(
        default=None,
        description="Thermal Conductivity, https://en.wikipedia.org/wiki/Thermal_conductivity_and_resistivity",
    )
    HollowSections: HollowSectionsRangeV1 | None = None
    HotRolledSections: HotRolledSectionsRangeV1 | None = None
    PlateSteel: PlateSteelRangeV1 | None = None


class PrefabricatedSteelAssembliesRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Prefabricated assemblies made primarily of steel.

    Range version.
    """

    _EXT_VERSION = "1.0"

    MetalRailings: MetalRailingsRangeV1 | None = None
    MetalStairs: MetalStairsRangeV1 | None = None
    MiscMetalFabrication: MiscMetalFabricationRangeV1 | None = None
    OpenWebMembranes: OpenWebMembranesRangeV1 | None = None


class PostTensioningSteelRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Steel tensioning cables or tendons for compression of prestressed concrete.

    Range version.
    """

    _EXT_VERSION = "1.0"


class RebarSteelRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Reinforcing bar used together with concrete.

    Range version.
    """

    _EXT_VERSION = "1.0"

    fabricated: bool | None = None
    grade: list[SteelRebarGrade] | None = pyd.Field(default=None)
    diameter_min: AmountRangeLengthMm | None = pyd.Field(default=None, description="Minimal diameter")
    bending_pin_max: RangeFloat | None = pyd.Field(default=None)
    ts_ys_ratio_max: RangeFloat | None = pyd.Field(
        default=None,
        description="Max ratio of ultimate tensile to yield tensile strength",
    )
    epoxy_coated: bool | None = pyd.Field(default=None)


class WireMeshSteelRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Mild steel wire for reinforcement, connections, and meshes.

    Range version.
    """

    _EXT_VERSION = "1.0"

    fabricated: bool | None = None


class OtherSteelRangeV1(BaseOpenEpdHierarchicalSpec):
    """Steel products that do not fit into a defined subcategory."""

    _EXT_VERSION = "1.0"


class SteelRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Broad category for construction materials made from steel and its alloys.

    Range version.
    """

    _EXT_VERSION = "1.1"

    yield_tensile_str: AmountRangePressureMpa | None = pyd.Field(
        default=None,
        description="Yield Tensile strength (Mpa) per unit area. Yield strength is the point at which a material begins to permanently deform or change shape due to applied stress.",
    )
    bar_elongation: RangeFloat | None = pyd.Field(
        default=None,
        description="Increase in length at break, in percent. Typically 10%-20%",
    )
    recycled_content: RangeRatioFloat | None = pyd.Field(default=None, description="")
    post_consumer_recycled_content: RangeRatioFloat | None = pyd.Field(
        default=None,
        description="Should be a number between zero and the Recycled Content (steel_recycled_content)",
    )
    astm_marking: str | None = pyd.Field(default=None, description="The marking to be expected on the product.")
    euro_marking: str | None = pyd.Field(default=None, description="The marking to be expected on the product.")
    composition: list[SteelComposition] | None = pyd.Field(default=None, description="Basic chemical composition")
    cold_finished: bool | None = pyd.Field(default=None)
    galvanized: bool | None = pyd.Field(default=None)
    stainless: bool | None = pyd.Field(default=None)
    making_route: SteelMakingRoute | None = pyd.Field(default=None)
    astm_standards: list[Standard] | None = pyd.Field(default=None, description="List of ASTM standards")
    sae_standards: list[Standard] | None = pyd.Field(default=None, description="List of SAE standards")
    en_standards: list[Standard] | None = pyd.Field(default=None, description="List of EN standards")
    MBQSteel: MBQSteelRangeV1 | None = None
    CoilSteel: CoilSteelRangeV1 | None = None
    ColdFormedSteel: ColdFormedSteelRangeV1 | None = None
    StructuralSteel: StructuralSteelRangeV1 | None = None
    PrefabricatedSteelAssemblies: PrefabricatedSteelAssembliesRangeV1 | None = None
    PostTensioningSteel: PostTensioningSteelRangeV1 | None = None
    RebarSteel: RebarSteelRangeV1 | None = None
    WireMeshSteel: WireMeshSteelRangeV1 | None = None
    OtherSteel: OtherSteelRangeV1 | None = None



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from typing import Annotated, Any, Callable, Type, TypeAlias

from openepd.compat.pydantic import pyd
from openepd.model.versioning import Version


def together_validator(field1: str, field2: Any, values: dict[str, Any]) -> Any:
    """Shared validator to ensure that two fields are provided together or not provided at all."""
    value1 = values.get(field1)
    value2 = values.get(field2)
    if value1 is not None and value2 is None or value1 is None and value2 is not None:
        raise ValueError(f"Both or neither {field1} and {field2} days must be provided together")


def validate_version_format(v: str) -> str:
    """Ensure that the extension version is valid."""
    Version.parse_version(v)  # will raise an error if not valid
    return v


def validate_version_compatibility(class_version_attribute_name: str) -> Callable[[Type, str], str]:
    """Ensure that the object which is passed for parsing and validation is compatible with the class."""

    # we need closure to pass property name, since actual class will only be available in runtime
    def internal_validate_version_compatibility(cls: Type, v: str) -> str:
        if not hasattr(cls, class_version_attribute_name):
            raise ValueError(f"Class {cls} must declare a class var extension var named {class_version_attribute_name}")

        class_version = getattr(cls, class_version_attribute_name)
        if Version.parse_version(v).major != Version.parse_version(class_version).major:
            raise ValueError(f"Extension version {v} does not match class version {class_version}")
        return v

    return internal_validate_version_compatibility


ReferenceStr: TypeAlias = Annotated[
    str,
    pyd.Field(description="Reference to another object", example="https://buildingtransparency.org/ec3/epds/1u7zsed8"),
]



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from openepd.api.dto.base import BaseMetaDto, BaseOpenEpdApiModel
from openepd.compat.pydantic import pyd


class PerformanceMeta(BaseMetaDto):
    """Meta for performance information."""

    execution_time_ms: int


class PerformanceMetaMixin(BaseMetaDto):
    """Mixin for adding performance meta to MetaCollection."""

    performance: PerformanceMeta | None = None


class PagingMeta(BaseMetaDto):
    """Meta for paging information."""

    total_count: int = pyd.Field(
        title="Total results", example=1233, description="Total number of records for the search"
    )
    total_pages: int = pyd.Field(title="Total pages", example=20, description="Total pages available")
    page_size: int = pyd.Field(title="Page size", example=150, description="Number of records in page")


class PagingMetaMixin(BaseOpenEpdApiModel):
    """Mixin for adding paging meta to MetaCollection."""

    paging: PagingMeta | None


class WarningMessageDto(BaseOpenEpdApiModel):
    """DTO for warning messages."""

    message: str = pyd.Field(
        title="Warning message", example="Categories limited during search, see effective_omf in meta"
    )
    code: str = pyd.Field(title="Warning code", example="CATEGORIES_LIMITED")
    field: str | None = pyd.Field(
        title="Field",
        description="Field to which the warning relates",
        example="subcategories",
    )


class WarningMetaMixin(BaseOpenEpdApiModel):
    """Mixin for adding warnings meta to MetaCollection."""

    warnings: list[WarningMessageDto] | None = None



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from openepd.compat.pydantic import pyd
from openepd.model.specs.base import BaseOpenEpdHierarchicalSpec
from openepd.model.validation.quantity import (
    PressureMPaStr,
    ThermalConductivityStr,
    validate_quantity_ge_factory,
    validate_quantity_unit_factory,
)


class GMUV1(BaseOpenEpdHierarchicalSpec):
    """Glass masonry unit."""

    _EXT_VERSION = "1.0"


class AutoclavedAeratedConcreteV1(BaseOpenEpdHierarchicalSpec):
    """A lightweight, precast, foamed concrete masonry building material."""

    _EXT_VERSION = "1.0"

    # Own fields:
    strength_28d: PressureMPaStr | None = pyd.Field(default=None, description="", example="1 MPa")
    thermal_conductivity: ThermalConductivityStr | None = pyd.Field(
        default=None, description="", example="1 W / (m * K)"
    )
    white: bool | None = pyd.Field(default=None, description="", example=True)

    _aac_thermal_conductivity_is_quantity_validator = pyd.validator("thermal_conductivity", allow_reuse=True)(
        validate_quantity_unit_factory("W / (m * K)")
    )

    _aac_thermal_conductivity_min_validator = pyd.validator("thermal_conductivity", allow_reuse=True)(
        validate_quantity_ge_factory("0 W / (m * K)")
    )


class BrickV1(BaseOpenEpdHierarchicalSpec):
    """Solid masonry units made from clay or shale."""

    _EXT_VERSION = "1.0"

    # Own fields:
    building: bool | None = pyd.Field(default=None, description="", example=True)
    facing: bool | None = pyd.Field(default=None, description="", example=True)
    floor: bool | None = pyd.Field(default=None, description="", example=True)
    pedestrian: bool | None = pyd.Field(default=None, description="", example=True)
    paving: bool | None = pyd.Field(default=None, description="", example=True)
    other: bool | None = pyd.Field(default=None, description="", example=True)
    chemical_resistant: bool | None = pyd.Field(default=None, description="", example=True)
    glazed: bool | None = pyd.Field(default=None, description="", example=True)
    tiles: bool | None = pyd.Field(default=None, description="", example=True)


class MasonryV1(BaseOpenEpdHierarchicalSpec):
    """Structural and/or enclosure system based on individual rigid units stacked and bound together with mortar."""

    _EXT_VERSION = "1.0"

    # Own fields:
    white_cement: bool | None = pyd.Field(default=None, description="", example=True)

    # Nested specs:
    GMU: GMUV1 | None = None
    AutoclavedAeratedConcrete: AutoclavedAeratedConcreteV1 | None = None
    Brick: BrickV1 | None = None



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from typing import Annotated

from openepd.compat.pydantic import pyd
from openepd.model.specs.base import BaseOpenEpdHierarchicalSpec, CodegenSpec
from openepd.model.specs.enums import AsphaltGradation, AsphaltMixType
from openepd.model.validation.numbers import RatioFloat
from openepd.model.validation.quantity import LengthMmStr, TemperatureCStr


class AsphaltV1(BaseOpenEpdHierarchicalSpec):
    """General category for asphalt mixtures."""

    _EXT_VERSION = "1.0"

    # Own fields:
    aggregate_size_max: LengthMmStr | None = pyd.Field(default=None, description="Max aggregate size", example="20 mm")
    rap: RatioFloat | None = pyd.Field(
        default=None,
        description="Percent of mixture that has been replaced by recycled asphalt pavement (RAP).",
        example=0.5,
        ge=0,
        le=1,
    )
    ras: RatioFloat | None = pyd.Field(
        default=None,
        description="Percent of mixture that has been replaced by recycled asphalt shingles (RAS).",
        example=0.5,
        ge=0,
        le=1,
    )
    ground_tire_rubber: RatioFloat | None = pyd.Field(
        default=None,
        description="Percent of mixture that has been replaced by ground tire rubber (GTR).",
        example=0.5,
        ge=0,
        le=1,
    )
    max_temperature: Annotated[TemperatureCStr | None, CodegenSpec(override_type=TemperatureCStr)] = pyd.Field(
        default=None,
        description="The upper threshold temperature to which an asphalt "
        "binder can be heated preventing the asphalt mixture "
        "from rutting",
        example="90 °C",
    )
    min_temperature: Annotated[TemperatureCStr | None, CodegenSpec(override_type=TemperatureCStr)] = pyd.Field(
        default=None,
        description="The lower threshold temperature for an asphalt "
        "binder to prevent thermal cracking of the asphalt"
        " mixture.",
        example="-20 °C",
    )
    mix_type: AsphaltMixType | None = pyd.Field(default=None, description="Asphalt mix type", example="WMA")
    gradation: AsphaltGradation | None = pyd.Field(default=None, description="Asphalt gradation", example="Gap-graded")

    sbr: bool | None = pyd.Field(default=None, description="Styrene-butadiene rubber (SBR)", example=True)
    sbs: bool | None = pyd.Field(default=None, description="Styrene-butadiene-styrene (SBS)", example=True)
    ppa: bool | None = pyd.Field(default=None, description="Polyphosphoric acid (PPA)", example=True)
    gtr: bool | None = pyd.Field(default=None, description="Ground tire rubber (GTR)", example=True)
    pmb: bool | None = pyd.Field(default=None, description="Polymer modified bitumen (PMB)", example=True)
